"""Defines several utility functions.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

from typing import Tuple

import jax
import jax.numpy as jnp
import numpy as onp
from jax.experimental import host_callback

EPS_EIG = 1e-6


def diag(x: jnp.ndarray) -> jnp.ndarray:
    """A batch-compatible version of `numpy.diag`."""
    shape = x.shape + (x.shape[-1],)
    y = jnp.zeros(shape, x.dtype)
    i = jnp.arange(x.shape[-1])
    return y.at[..., i, i].set(x)


def angular_frequency_for_wavelength(wavelength: jnp.ndarray) -> jnp.ndarray:
    """Returns the angular frequency for the specified wavelength."""
    return 2 * jnp.pi / wavelength  # Since by our convention c == 1.


def matrix_adjoint(x: jnp.ndarray) -> jnp.ndarray:
    """Computes the adjoint for a batch of matrices."""
    axes = tuple(range(x.ndim - 2)) + (x.ndim - 1, x.ndim - 2)
    return jnp.conj(jnp.transpose(x, axes=axes))


def batch_compatible_shapes(*shapes: Tuple[int, ...]) -> bool:
    """Returns `True` if all the shapes are batch-compatible."""
    max_dims = max([len(s) for s in shapes])
    shapes = tuple([(1,) * (max_dims - len(s)) + s for s in shapes])
    max_shape = [max(dim_shapes) for dim_shapes in zip(*shapes)]
    for shape in shapes:
        if any([a not in (1, b) for a, b in zip(shape, max_shape)]):
            return False
    return True


def atleast_nd(x: jnp.ndarray, n: int) -> jnp.ndarray:
    """Adds leading dimensions to `x`, ensuring that it is at least n-dimensional."""
    dims_to_add = tuple(range(max(0, n - x.ndim)))
    return jnp.expand_dims(x, axis=dims_to_add)


def absolute_axes(axes: Tuple[int, ...], ndim: int) -> Tuple[int, ...]:
    """Returns the absolute axes for given relative axes and number of array dimensions."""
    if not all(a in list(range(-ndim, ndim)) for a in axes):
        raise ValueError(
            f"All elements of `axes` must be in the range ({ndim}, {ndim - 1}) "
            f"but got {axes}."
        )
    absolute_axes = tuple([d % ndim for d in axes])
    if len(absolute_axes) != len(set(absolute_axes)):
        raise ValueError(
            f"Found duplicates in `axes`; computed absolute axes are {absolute_axes}."
        )
    return absolute_axes


def interpolate_permittivity(
    permittivity_solid: jnp.ndarray,
    permittivity_void: jnp.ndarray,
    density: jnp.ndarray,
) -> jnp.ndarray:
    """Interpolates the permittivity with a scheme that avoids zero crossings.

    The interpolation uses the scheme introduced in [2019 Christiansen], which avoids
    zero crossings that can occur with metals or lossy materials having a negative
    real component of the permittivity. https://doi.org/10.1016/j.cma.2018.08.034

    Args:
        permittivity_solid: The permittivity of solid regions.
        permittivity_void: The permittivity of void regions.
        density: The density, specifying which locations are solid and which are void.

    Returns:
        The interpolated permittivity.
    """
    n_solid = jnp.real(jnp.sqrt(permittivity_solid))
    k_solid = jnp.imag(jnp.sqrt(permittivity_solid))
    n_void = jnp.real(jnp.sqrt(permittivity_void))
    k_void = jnp.imag(jnp.sqrt(permittivity_void))
    n = density * n_solid + (1 - density) * n_void
    k = density * k_solid + (1 - density) * k_void
    return (n + 1j * k) ** 2


def magnitude(tx: jnp.ndarray, ty: jnp.ndarray) -> jnp.ndarray:
    """Computes elementwise magnitude of the vector field defined by `(tx, ty)`.

    This method computes the magnitude with special logic that avoides `nan`
    gradients when the magnitude is zero.

    Args:
        tx: Array giving the x-component of the vector field.
        ty: Array giving the y-component of the vector field.

    Returns:
        Array giving the vector magnitude.
    """
    # Avoid taking the square root of an array with any zero elements, to avoid
    # `nan` in the gradients.
    magnitude_squared = jnp.abs(tx) ** 2 + jnp.abs(ty) ** 2
    is_zero = magnitude_squared == 0
    magnitude_squared_safe = jnp.where(is_zero, 1.0, magnitude_squared)
    return jnp.where(is_zero, 0.0, jnp.sqrt(magnitude_squared_safe))


def angle(x: jnp.ndarray) -> jnp.ndarray:
    """Computes `angle(x)` with special logic for when `x` equals zero."""
    # Avoid taking the angle of an array with any near-zero elements, to avoid
    # `nan` in the gradients.
    is_near_zero = jnp.isclose(x, 0.0)
    x_safe = jnp.where(is_near_zero, (1.0 + 0.0j), x)
    return jnp.angle(x_safe)


def resample(
    x: jnp.ndarray,
    shape: Tuple[int, ...],
    method=jax.image.ResizeMethod,
) -> jnp.ndarray:
    """Resamples `x` to have the specified `shape`.

    The algorithm first upsamples `x` so that the pixels in the output image are
    comprised of an integer number of pixels in the upsampled `x`, and then
    performs box downsampling.

    Args:
        x: The array to be resampled.
        shape: The shape of the output array.
        method: The method used to resize `x` prior to box downsampling.

    Returns:
        The resampled array.
    """
    if x.ndim != len(shape):
        raise ValueError(
            f"`shape` must have length matching number of dimensions in `x`, "
            f"but got {shape} when `x` had shape {x.shape}."
        )

    factor = [int(onp.ceil(dx / d)) for dx, d in zip(x.shape, shape)]
    upsampled_shape = tuple([d * f for d, f in zip(shape, factor)])

    x_upsampled = jax.image.resize(
        image=x,
        shape=upsampled_shape,
        method=method,
    )

    return box_downsample(x_upsampled, shape)


def box_downsample(x: jnp.ndarray, shape: Tuple[int, ...]) -> jnp.ndarray:
    """Downsamples `x` to a coarser resolution array using box downsampling.

    Box downsampling forms nonoverlapping windows and simply averages the
    pixels within each window. For example, downsampling `(0, 1, 2, 3, 4, 5)`
    with a factor of `2` yields `(0.5, 2.5, 4.5)`.

    Args:
        x: The array to be downsampled.
        shape: The shape of the output array; each axis dimension must evenly
            divide the corresponding axis dimension in `x`.

    Returns:
        The output array with shape `shape`.
    """
    if x.ndim != len(shape) or any([(d % s) != 0 for d, s in zip(x.shape, shape)]):
        raise ValueError(
            f"Each axis of `shape` must evenly divide the corresponding axis "
            f"dimension in `x`, but got shape {shape} when `x` has shape "
            f"{x.shape}."
        )
    shape = sum([(s, d // s) for d, s in zip(x.shape, shape)], ())
    axes = list(range(1, 2 * x.ndim, 2))
    x = x.reshape(shape)
    return jnp.mean(x, axis=axes)


# -----------------------------------------------------------------------------
# Functions related to convolutions and kernels.
# -----------------------------------------------------------------------------


def padded_conv(
    x: jnp.ndarray,
    kernel: jnp.ndarray,
    padding_mode: str,
) -> jnp.ndarray:
    """Convolves `x` with `kernel`, using padding with the specified mode.

    Before the convolution, `x` is padded using the specified padding mode.

    Args:
        x: The source array.
        kernel: The rank-2 convolutional kernel.
        padding_mode: One of the padding modes supported by `jnp.pad`.

    Returns:
        The result of the convolution.
    """
    assert kernel.ndim == 2
    pad_ij = [(s // 2 - (s + 1) % 2, s // 2) for s in kernel.shape]
    batch_dims = x.ndim - 2
    batch_shape = x.shape[:-2]
    padding = [(0, 0)] * batch_dims + pad_ij
    x = jnp.pad(x, padding, mode=padding_mode)
    # Flatten the batch dimensions and add a dummy channel dimension.
    x = x.reshape((-1,) + x.shape[-2:])
    x = jnp.expand_dims(x, -3)
    y = jax.lax.conv_general_dilated(
        x,
        kernel[jnp.newaxis, jnp.newaxis, :, :].astype(x.dtype),
        window_strides=(1, 1),
        padding="VALID",
        dimension_numbers=("NCHW", "OIHW", "NCHW"),
    )
    # Remove the dummy channel dimension, and restore the original batch shape.
    y = jnp.squeeze(y, axis=-3)
    y = y.reshape(batch_shape + y.shape[-2:])
    return y


def gaussian_kernel(shape: Tuple[int, int], fwhm: float) -> jnp.ndarray:
    """Generates a Gaussian kernel with the specified shape."""
    x = jnp.arange(shape[0])[:, jnp.newaxis]
    y = jnp.arange(shape[1])[jnp.newaxis, :]
    # Ensure that the center is on a pixel, not between pixels.
    x_center = (shape[0] - 0.5) // 2
    y_center = (shape[1] - 0.5) // 2
    distance_squared = (x - x_center) ** 2 + (y - y_center) ** 2
    sigma = fwhm / (2 * jnp.sqrt(2 * jnp.log(2)))

    # When `fwhm` is zero, we want to return a one-hot array.
    fwhm_is_zero = fwhm == 0.0
    denom_safe = jnp.where(fwhm_is_zero, 1.0, 2 * sigma**2)
    return jnp.where(
        fwhm_is_zero,
        (distance_squared == 0).astype(float),
        jnp.exp(-distance_squared / denom_safe),
    )


# -----------------------------------------------------------------------------
# Functions related to a generalized eigensolve with custom vjp rule.
# -----------------------------------------------------------------------------


@jax.custom_vjp
def eig(matrix: jnp.ndarray, eps: float = EPS_EIG) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Wraps `jnp.linalg.eig` in a jit-compatible, differentiable manner.

    The custom vjp allows gradients with resepct to the eigenvectors, unlike the
    standard jax implementation of `eig`. We use an expression for the gradient
    given in [2019 Boeddeker] along with a regularization scheme used in [2021
    Colburn]. The method effectively applies a Lorentzian broadening to a term
    containing the inverse difference of eigenvalues.

    [2019 Boeddeker] https://arxiv.org/abs/1701.00392
    [2021 Coluburn] https://www.nature.com/articles/s42005-021-00568-6

    Args:
        matrix: The matrix for which eigenvalues and eigenvectors are sought.
        eps: Parameter which determines the degree of broadening.

    Returns:
        The eigenvalues and eigenvectors.
    """
    del eps
    return _eig_host(matrix)


def _eig_host(matrix: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Wraps jnp.linalg.eig so that it can be jit-ed on a machine with GPUs."""
    eigenvalues_shape = jax.ShapeDtypeStruct(matrix.shape[:-1], complex)
    eigenvectors_shape = jax.ShapeDtypeStruct(matrix.shape, complex)

    def _eig_cpu(matrix: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray]:
        # We force this computation to be performed on the cpu by jit-ing and
        # explicitly specifying the device.
        with jax.default_device(jax.devices("cpu")[0]):
            return jax.jit(jnp.linalg.eig)(matrix)

    return host_callback.call(
        _eig_cpu,
        matrix.astype(complex),
        result_shape=(eigenvalues_shape, eigenvectors_shape),
    )


def _eig_fwd(
    matrix: jnp.ndarray,
    eps: float,
) -> Tuple[Tuple[jnp.ndarray, jnp.ndarray], Tuple[jnp.ndarray, jnp.ndarray, float]]:
    """Implements the forward calculation for `eig`."""
    eigenvalues, eigenvectors = _eig_host(matrix)
    return (eigenvalues, eigenvectors), (eigenvalues, eigenvectors, eps)


def _eig_bwd(
    res: Tuple[jnp.ndarray, jnp.ndarray, float],
    grads: Tuple[jnp.ndarray, jnp.ndarray],
) -> Tuple[jnp.ndarray, None]:
    """Implements the backward calculation for `eig`."""
    eigenvalues, eigenvectors, eps = res
    grad_eigenvalues, grad_eigenvectors = grads

    # Compute the F-matrix, from equation 5 of [2021 Colburn]. This applies a
    # Lorentzian broadening to the matrix `f = 1 / (eigenvalues[i] - eigenvalues[j])`.
    eigenvalues_i = eigenvalues[..., jnp.newaxis, :]
    eigenvalues_j = eigenvalues[..., :, jnp.newaxis]
    f_broadened = (eigenvalues_i - eigenvalues_j) / (
        (eigenvalues_i - eigenvalues_j) ** 2 + eps
    )

    # Manually set the diagonal elements to zero, as we do not use broadening here.
    i = jnp.arange(f_broadened.shape[-1])
    f_broadened = f_broadened.at[..., i, i].set(0)

    # By jax convention, gradients are with respect to the complex parameters, not with
    # respect to their conjugates. Take the conjugates.
    grad_eigenvalues_conj = jnp.conj(grad_eigenvalues)
    grad_eigenvectors_conj = jnp.conj(grad_eigenvectors)

    eigenvectors_H = matrix_adjoint(eigenvectors)
    dim = eigenvalues.shape[-1]
    eye_mask = jnp.eye(dim, dtype=bool)
    eye_mask = eye_mask.reshape((1,) * (eigenvalues.ndim - 1) + (dim, dim))

    # Then, the gradient is found by equation 4.77 of [2019 Boeddeker].
    rhs = (
        diag(grad_eigenvalues_conj)
        + jnp.conj(f_broadened) * (eigenvectors_H @ grad_eigenvectors_conj)
        - jnp.conj(f_broadened)
        * (eigenvectors_H @ eigenvectors)
        @ jnp.where(eye_mask, jnp.real(eigenvectors_H @ grad_eigenvectors_conj), 0.0)
    ) @ eigenvectors_H
    grad_matrix = jnp.linalg.solve(eigenvectors_H, rhs)

    # Take the conjugate of the gradient, reverting to the jax convention
    # where gradients are with respect to complex parameters.
    grad_matrix = jnp.conj(grad_matrix)

    # Return `grad_matrix`, and `None` for the gradient with respect to `eps`.
    return grad_matrix, None


eig.defvjp(_eig_fwd, _eig_bwd)



"""Tests for `examples.uled`.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

import unittest

import jax.numpy as jnp
import numpy as onp

from examples import uled
from fmmax import basis, fmm

SIM_CONFIG_KWARGS = {
    "formulation": fmm.Formulation.JONES_DIRECT,
    "truncation": basis.Truncation.PARALLELOGRAMIC,
    "approximate_num_terms": 100,
}


class MicroLedTest(unittest.TestCase):
    def test_regression(self):
        # Checks that results match reference values. This helps protect against
        # regressions in accuracy of the simulator.
        extraction_efficiency, _, efields, hfields, _ = uled.simulate_uled(
            resolution=25,
            resolution_fields=25,
            dipole_fwhm=0,
            dipole_y_offset=(0,),
            **SIM_CONFIG_KWARGS,
        )
        with self.subTest("extraction efficiency"):
            onp.testing.assert_allclose(
                extraction_efficiency,
                [0.494347, 0.492305, 0.416735],
                atol=1e-3,
            )

        with self.subTest("efields"):
            self.assertSequenceEqual(efields.shape, (3, 1, 1, 56, 56, 48, 3))
            onp.testing.assert_allclose(
                jnp.mean(jnp.abs(efields) ** 2, axis=(1, 2, 3, 4, 5)),
                [
                    [64.444786, 16.373577, 8.459116],
                    [16.909363, 64.59829, 8.159439],
                    [28.255959, 28.230724, 8.761396],
                ],
                rtol=1e-3,
            )

        with self.subTest("hfields"):
            self.assertSequenceEqual(hfields.shape, (3, 1, 1, 56, 56, 48, 3))
            onp.testing.assert_allclose(
                jnp.mean(jnp.abs(hfields) ** 2, axis=(1, 2, 3, 4, 5)),
                [
                    [92.3925, 428.87042, 46.39356],
                    [427.06384, 90.633675, 48.09156],
                    [115.1582, 114.41496, 28.075932],
                ],
                rtol=1e-3,
            )



"""Tests for `examples.anisotropic_grating`.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

import unittest

import jax.numpy as jnp
import numpy as onp

from examples import anisotropic_grating


class AnisotropicGratingTest(unittest.TestCase):
    def test_regression(self):
        # Checks that results match reference values. This helps protect against
        # regressions in accuracy of the simulator.
        _, incident, reflected, transmitted = anisotropic_grating.simulate_grating(
            wavelength_nm=jnp.array([450.0, 550.0, 620.0]),
            polar_angle=jnp.array([[0.0], [0.0], [0.1], [0.1]]),
            azimuthal_angle=jnp.array([[0.0], [0.1], [0.0], [0.1]]),
        )

        # Sum over the orders, giving the per-wavelength and per-polarization values.
        incident = jnp.sum(incident, axis=-2)
        reflected = jnp.sum(reflected, axis=-2)
        transmitted = jnp.sum(transmitted, axis=-2)

        onp.testing.assert_allclose(-reflected + transmitted, incident, rtol=1e-5)



"""Tests for `fmmax.fields`.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

import dataclasses
import unittest

import jax
import jax.numpy as jnp
import numpy as onp
import parameterized

from fmmax import basis, fields, fmm, layer, scattering


def example_solve(permittivity_batch_shape, wavelength_batch_shape):
    # Solves for the interior scattering matrices of a simple structure.
    random_density = jax.random.uniform(
        jax.random.PRNGKey(0), permittivity_batch_shape + (20, 20)
    )
    permittivities = [
        jnp.ones(permittivity_batch_shape + (1, 1), dtype=complex),
        jnp.ones(permittivity_batch_shape + (1, 1), dtype=complex) * 2.25,
        random_density + 2,
        jnp.ones(permittivity_batch_shape + (1, 1), dtype=complex),
    ]
    thicknesses = [10, 10, 10, 10]
    in_plane_wavevector = jnp.zeros((2,))
    primitive_lattice_vectors = basis.LatticeVectors(u=basis.X, v=basis.Y)
    expansion = basis.generate_expansion(
        primitive_lattice_vectors=primitive_lattice_vectors,
        approximate_num_terms=10,
        truncation=basis.Truncation.CIRCULAR,
    )
    layer_solve_results = [
        layer.eigensolve_isotropic_media(
            wavelength=jnp.full(wavelength_batch_shape, 100.0),
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=primitive_lattice_vectors,
            permittivity=p,
            expansion=expansion,
            formulation=fmm.Formulation.FFT,
        )
        for p in permittivities
    ]

    s_matrices_interior = scattering.stack_s_matrices_interior(
        layer_solve_results, thicknesses
    )
    return layer_solve_results, thicknesses, s_matrices_interior


def time_average_z_poynting_flux(electric_fields, magnetic_fields):
    ex, ey, _ = electric_fields
    hx, hy, _ = magnetic_fields
    smz = ex * jnp.conj(hy) - ey * jnp.conj(hx)
    return jnp.real(smz)


class ShapesTest(unittest.TestCase):
    @parameterized.parameterized.expand(
        [
            # permittivity, wavelength, and excitation batch shape.
            [(), (), (1,)],
            [(3,), (), (1,)],
            [(), (), (2,)],
            [(3,), (), (2,)],
            [(2, 1, 3), (), (2,)],
            [(), (3,), (1,)],
            [(3,), (2, 1), (1,)],
            [(), (3,), (2,)],
            [(3,), (2, 1), (2,)],
            [(2, 1, 3), (2, 1, 1), (2,)],
        ]
    )
    def test_shapes_match_expected(
        self, permittivity_batch_shape, wavelength_batch_shape, excitation_batch_shape
    ):
        # Tests various functions in the `fields` module, validating compatibility
        # with various batch sizes for the excitation and simulation.
        layer_solve_results, thicknesses, s_matrices_interior = example_solve(
            permittivity_batch_shape, wavelength_batch_shape
        )

        num_terms = layer_solve_results[0].expansion.num_terms
        forward_amplitude_0_start = jnp.zeros(
            (2 * num_terms,) + excitation_batch_shape, dtype=complex
        )
        forward_amplitude_0_start = forward_amplitude_0_start.at[0, :].set(1)
        backward_amplitude_N_end = jnp.zeros_like(forward_amplitude_0_start)

        ab = fields.stack_amplitudes_interior(
            s_matrices_interior, forward_amplitude_0_start, backward_amplitude_N_end
        )

        for (
            (forward_amplitude_start, backward_amplitude_end),
            layer_solve_result,
            thickness,
        ) in zip(ab, layer_solve_results, thicknesses):
            a, b = fields.colocate_amplitudes(
                forward_amplitude_start,
                backward_amplitude_end,
                z_offset=thickness / 2,
                layer_solve_result=layer_solve_result,
                layer_thickness=thickness,
            )
            batch_shape = jnp.broadcast_shapes(
                permittivity_batch_shape, wavelength_batch_shape
            )
            expected_shape = batch_shape + (num_terms * 2,) + excitation_batch_shape
            self.assertSequenceEqual(a.shape, expected_shape)
            self.assertSequenceEqual(b.shape, expected_shape)

            # Compute the electric and magnetic field Fourier coefficeints.
            ef, hf = fields.fields_from_wave_amplitudes(
                a, b, layer_solve_result=layer_solve_result
            )
            ex, ey, ez = ef
            hx, hy, hz = hf
            expected_shape = batch_shape + (num_terms,) + excitation_batch_shape
            self.assertSequenceEqual(ex.shape, expected_shape)
            self.assertSequenceEqual(ey.shape, expected_shape)
            self.assertSequenceEqual(ez.shape, expected_shape)
            self.assertSequenceEqual(hx.shape, expected_shape)
            self.assertSequenceEqual(hy.shape, expected_shape)
            self.assertSequenceEqual(hz.shape, expected_shape)

            # Compute the electric and magnetic fields on the grid.
            grid_shape = (10, 11)
            eg, hg, (x, y) = fields.fields_on_grid(
                ef,
                hf,
                layer_solve_result=layer_solve_result,
                shape=grid_shape,
                num_unit_cells=(1, 1),
            )
            egx, egy, egz = eg
            hgx, hgy, hgz = hg
            expected_shape = batch_shape + grid_shape + excitation_batch_shape
            self.assertSequenceEqual(egx.shape, expected_shape)
            self.assertSequenceEqual(egy.shape, expected_shape)
            self.assertSequenceEqual(egz.shape, expected_shape)
            self.assertSequenceEqual(hgx.shape, expected_shape)
            self.assertSequenceEqual(hgy.shape, expected_shape)
            self.assertSequenceEqual(hgz.shape, expected_shape)
            self.assertSequenceEqual(x.shape, grid_shape)
            self.assertSequenceEqual(y.shape, grid_shape)


class FieldsOnCoordinatesTest(unittest.TestCase):
    @parameterized.parameterized.expand([[()], [(3,)]])
    def test_fields_on_coordinates_match_fields_on_grid(self, batch_shape):
        in_plane_wavevector = jnp.zeros((2,))
        primitive_lattice_vectors = basis.LatticeVectors(u=basis.X, v=basis.Y)
        expansion = basis.generate_expansion(
            primitive_lattice_vectors=primitive_lattice_vectors,
            approximate_num_terms=10,
            truncation=basis.Truncation.CIRCULAR,
        )

        permittivity = 1 + jax.random.uniform(
            jax.random.PRNGKey(0), batch_shape + (20, 20)
        )
        layer_solve_result = layer.eigensolve_isotropic_media(
            wavelength=jnp.ones(()),
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=primitive_lattice_vectors,
            permittivity=permittivity,
            expansion=expansion,
            formulation=fmm.Formulation.FFT,
        )

        ex, ey, ez, hx, hy, hz = jax.random.normal(
            jax.random.PRNGKey(0),
            shape=(6,) + batch_shape + (layer_solve_result.expansion.num_terms, 5),
            dtype=complex,
        )
        expected_efield, expected_hfield, (x, y) = fields.fields_on_grid(
            electric_field=(ex, ey, ez),
            magnetic_field=(hx, hy, hz),
            layer_solve_result=layer_solve_result,
            shape=(20, 20),
            num_unit_cells=(1, 1),
        )

        with self.subTest("xy on a square grid"):
            efield, hfield, _ = fields.fields_on_coordinates(
                electric_field=(ex, ey, ez),
                magnetic_field=(hx, hy, hz),
                layer_solve_result=layer_solve_result,
                x=x,
                y=y,
            )
            onp.testing.assert_allclose(efield, expected_efield, rtol=1e-4)
            onp.testing.assert_allclose(hfield, expected_hfield, rtol=1e-4)

        with self.subTest("xy as flat arrays"):
            efield, hfield, _ = fields.fields_on_coordinates(
                electric_field=(ex, ey, ez),
                magnetic_field=(hx, hy, hz),
                layer_solve_result=layer_solve_result,
                x=x.flatten(),
                y=y.flatten(),
            )
            efield = onp.reshape(efield, onp.shape(expected_efield))
            hfield = onp.reshape(hfield, onp.shape(expected_efield))
            onp.testing.assert_allclose(efield, expected_efield, rtol=1e-4)
            onp.testing.assert_allclose(hfield, expected_hfield, rtol=1e-4)

    @parameterized.parameterized.expand([[()], [(3,)]])
    def test_layer_fields_on_coordinates_match_fields_on_grid(self, batch_shape):
        in_plane_wavevector = jnp.zeros((2,))
        primitive_lattice_vectors = basis.LatticeVectors(u=basis.X, v=basis.Y)
        expansion = basis.generate_expansion(
            primitive_lattice_vectors=primitive_lattice_vectors,
            approximate_num_terms=10,
            truncation=basis.Truncation.CIRCULAR,
        )

        permittivity = 1 + jax.random.uniform(
            jax.random.PRNGKey(0), batch_shape + (20, 20)
        )
        layer_solve_result = layer.eigensolve_isotropic_media(
            wavelength=jnp.ones(()),
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=primitive_lattice_vectors,
            permittivity=permittivity,
            expansion=expansion,
            formulation=fmm.Formulation.FFT,
        )

        fwd_amplitude_start, bwd_amplitude_end = jax.random.normal(
            jax.random.PRNGKey(0),
            shape=(2,) + batch_shape + (layer_solve_result.expansion.num_terms * 2, 5),
            dtype=complex,
        )
        expected_efield, expected_hfield, (x, y, _) = fields.layer_fields_3d(
            forward_amplitude_start=fwd_amplitude_start,
            backward_amplitude_end=bwd_amplitude_end,
            layer_solve_result=layer_solve_result,
            layer_thickness=jnp.ones(()),
            layer_znum=10,
            grid_shape=(20, 20),
            num_unit_cells=(1, 1),
        )

        with self.subTest("xy on a square grid"):
            efield, hfield, _ = fields.layer_fields_3d_on_coordinates(
                forward_amplitude_start=fwd_amplitude_start,
                backward_amplitude_end=bwd_amplitude_end,
                layer_solve_result=layer_solve_result,
                layer_thickness=jnp.ones(()),
                layer_znum=10,
                x=x,
                y=y,
            )
            onp.testing.assert_allclose(efield, expected_efield, rtol=1e-4)
            onp.testing.assert_allclose(hfield, expected_hfield, rtol=1e-4)

        with self.subTest("xy as flat arrays"):
            efield, hfield, _ = fields.layer_fields_3d_on_coordinates(
                forward_amplitude_start=fwd_amplitude_start,
                backward_amplitude_end=bwd_amplitude_end,
                layer_solve_result=layer_solve_result,
                layer_thickness=jnp.ones(()),
                layer_znum=10,
                x=x.flatten(),
                y=y.flatten(),
            )
            efield = onp.reshape(efield, onp.shape(expected_efield))
            hfield = onp.reshape(hfield, onp.shape(expected_efield))
            onp.testing.assert_allclose(efield, expected_efield, rtol=1e-4)
            onp.testing.assert_allclose(hfield, expected_hfield, rtol=1e-4)

    @parameterized.parameterized.expand([[()], [(3,)]])
    def test_stack_fields_on_coordinates_match_fields_on_grid(self, batch_shape):
        layer_solve_results, thicknesses, s_matrices_interior = example_solve(
            batch_shape, ()
        )
        num_terms = layer_solve_results[0].expansion.num_terms
        forward_amplitude_0_start = jax.random.normal(
            jax.random.PRNGKey(0),
            shape=batch_shape + (2 * num_terms, 1),
            dtype=complex,
        )
        backward_amplitude_N_end = jnp.zeros_like(forward_amplitude_0_start)
        ab = fields.stack_amplitudes_interior(
            s_matrices_interior, forward_amplitude_0_start, backward_amplitude_N_end
        )
        expected_efield, expected_hfield, (x, y, z) = fields.stack_fields_3d(
            amplitudes_interior=ab,
            layer_solve_results=layer_solve_results,
            layer_thicknesses=thicknesses,
            layer_znum=[10] * len(thicknesses),
            grid_shape=(10, 10),
            num_unit_cells=(1, 1),
        )
        efield, hfield, (x, y, z) = fields.stack_fields_3d_on_coordinates(
            amplitudes_interior=ab,
            layer_solve_results=layer_solve_results,
            layer_thicknesses=thicknesses,
            layer_znum=[10] * len(thicknesses),
            x=x,
            y=y,
        )
        onp.testing.assert_allclose(efield, expected_efield, rtol=1e-4)
        onp.testing.assert_allclose(hfield, expected_hfield, rtol=1e-4)


class PoyntingFluxTest(unittest.TestCase):
    @parameterized.parameterized.expand(
        [
            [(), ()],
            [(3,), ()],
            [(), ()],
            [(3,), ()],
            [(2, 1, 3), ()],
            [(), (3,)],
            [(3,), (2, 1)],
            [(2, 1, 3), (2, 1, 1)],
        ]
    )
    def test_eigenmode_poynting_flux(
        self, permittivity_batch_shape, wavelength_batch_shape
    ):
        layer_solve_results, _, _ = example_solve(
            permittivity_batch_shape, wavelength_batch_shape
        )
        layer_solve_result = layer_solve_results[0]
        eigenmode_flux = fields.eigenmode_poynting_flux(layer_solve_result)
        num_eigenmodes = layer_solve_result.eigenvalues.shape[-1]
        one_hot_amplitude = jnp.eye(num_eigenmodes)
        expected_eigenmode_flux, _ = fields.amplitude_poynting_flux(
            forward_amplitude=one_hot_amplitude,
            backward_amplitude=jnp.zeros_like(one_hot_amplitude),
            layer_solve_result=layer_solve_result,
        )
        expected_eigenmode_flux = jnp.sum(expected_eigenmode_flux, axis=-2)
        onp.testing.assert_allclose(eigenmode_flux, expected_eigenmode_flux)

    @parameterized.parameterized.expand(
        [
            [(), (), (1,)],
            [(3,), (), (1,)],
            [(), (), (2,)],
            [(3,), (), (2,)],
            [(2, 1, 3), (), (2,)],
            [(), (3,), (1,)],
            [(3,), (2, 1), (1,)],
            [(), (3,), (2,)],
            [(3,), (2, 1), (2,)],
            [(2, 1, 3), (2, 1, 1), (2,)],
        ]
    )
    def test_amplitude_poynting_flux(
        self, permittivity_batch_shape, wavelength_batch_shape, excitation_batch_shape
    ):
        layer_solve_results, thicknesses, s_matrices_interior = example_solve(
            permittivity_batch_shape, wavelength_batch_shape
        )

        num_eigenmodes = layer_solve_results[0].eigenvalues.shape[-1]
        amplitude_shape = (num_eigenmodes,) + excitation_batch_shape
        forward_amplitude_0_start = jax.random.uniform(
            jax.random.PRNGKey(0), amplitude_shape
        ).astype(complex)
        backward_amplitude_N_end = jnp.zeros_like(forward_amplitude_0_start)
        ab = fields.stack_amplitudes_interior(
            s_matrices_interior, forward_amplitude_0_start, backward_amplitude_N_end
        )

        for (
            (forward_amplitude_start, backward_amplitude_end),
            layer_solve_result,
            thickness,
        ) in zip(ab, layer_solve_results, thicknesses):
            a, b = fields.colocate_amplitudes(
                forward_amplitude_start,
                backward_amplitude_end,
                z_offset=thickness / 2,
                layer_solve_result=layer_solve_result,
                layer_thickness=thickness,
            )
            ef, hf = fields.fields_from_wave_amplitudes(
                a, b, layer_solve_result=layer_solve_result
            )
            grid_shape = (10, 11)
            eg, hg, _ = fields.fields_on_grid(
                ef,
                hf,
                layer_solve_result=layer_solve_result,
                shape=grid_shape,
                num_unit_cells=(1, 1),
            )
            sgz = time_average_z_poynting_flux(eg, hg)
            s_forward, s_backward = fields.amplitude_poynting_flux(
                a, b, layer_solve_result
            )
            onp.testing.assert_allclose(
                jnp.sum(s_forward + s_backward, axis=-2),
                jnp.mean(sgz, axis=(-3, -2)),
                rtol=1e-5,
            )

    @parameterized.parameterized.expand(
        [
            [(), (), (1,)],
            [(3,), (), (1,)],
            [(), (), (2,)],
            [(3,), (), (2,)],
            [(2, 1, 3), (), (2,)],
            [(), (3,), (1,)],
            [(3,), (2, 1), (1,)],
            [(), (3,), (2,)],
            [(3,), (2, 1), (2,)],
            [(2, 1, 3), (2, 1, 1), (2,)],
        ]
    )
    def test_directional_poynting_flux(
        self, permittivity_batch_shape, wavelength_batch_shape, excitation_batch_shape
    ):
        layer_solve_results, thicknesses, s_matrices_interior = example_solve(
            permittivity_batch_shape, wavelength_batch_shape
        )

        num_eigenmodes = layer_solve_results[0].eigenvalues.shape[-1]
        amplitude_shape = (num_eigenmodes,) + excitation_batch_shape
        forward_amplitude_0_start = jax.random.uniform(
            jax.random.PRNGKey(0), amplitude_shape
        ).astype(complex)
        backward_amplitude_N_end = jnp.zeros_like(forward_amplitude_0_start)
        ab = fields.stack_amplitudes_interior(
            s_matrices_interior, forward_amplitude_0_start, backward_amplitude_N_end
        )

        for (
            (forward_amplitude_start, backward_amplitude_end),
            layer_solve_result,
            thickness,
        ) in zip(ab, layer_solve_results, thicknesses):
            a, b = fields.colocate_amplitudes(
                forward_amplitude_start,
                backward_amplitude_end,
                z_offset=thickness / 2,
                layer_solve_result=layer_solve_result,
                layer_thickness=thickness,
            )
            ef, hf = fields.fields_from_wave_amplitudes(
                a, b, layer_solve_result=layer_solve_result
            )
            grid_shape = (10, 11)
            eg, hg, _ = fields.fields_on_grid(
                ef,
                hf,
                layer_solve_result=layer_solve_result,
                shape=grid_shape,
                num_unit_cells=(1, 1),
            )
            sgz = time_average_z_poynting_flux(eg, hg)
            s_forward, s_backward = fields.directional_poynting_flux(
                a, b, layer_solve_result
            )
            onp.testing.assert_allclose(
                jnp.sum(s_forward + s_backward, axis=-2),
                jnp.mean(sgz, axis=(-3, -2)),
                rtol=1e-5,
            )

            # Randomly flip the sign of some eigenvalues, and check that the
            # directional Poynting flux is unchanged.
            q = layer_solve_result.eigenvalues
            mask = jax.random.uniform(jax.random.PRNGKey(0), (q.shape[-1],))
            mask = jnp.ones_like(mask)
            a_swapped = jnp.where(mask[:, jnp.newaxis], a, b)
            b_swapped = jnp.where(mask[:, jnp.newaxis], b, a)
            layer_solve_result_swapped = dataclasses.replace(
                layer_solve_result, eigenvalues=jnp.where(mask, q, -q)
            )
            s_forward_swapped, s_backward_swapped = fields.directional_poynting_flux(
                a_swapped, b_swapped, layer_solve_result_swapped
            )

            onp.testing.assert_allclose(s_forward, s_forward_swapped)
            onp.testing.assert_allclose(s_backward, s_backward_swapped)
            onp.testing.assert_allclose(
                jnp.sum(s_forward_swapped + s_backward_swapped, axis=-2),
                jnp.mean(sgz, axis=(-3, -2)),
                rtol=1e-5,
            )


class Fields3DTest(unittest.TestCase):
    @parameterized.parameterized.expand(
        [
            # permittivity, wavelength, and excitation batch shape.
            [(), (), (1,), (1, 1)],
            [(3,), (), (1,), (1, 1)],
            [(), (), (2,), (1, 1)],
            [(3,), (), (2,), (1, 1)],
            [(2, 1, 3), (), (2,), (1, 1)],
            [(), (3,), (1,), (1, 1)],
            [(3,), (2, 1), (1,), (1, 1)],
            [(), (3,), (2,), (1, 1)],
            [(3,), (2, 1), (2,), (1, 1)],
            [(2, 1, 3), (2, 1, 1), (2,), (1, 1)],
            [(2, 1, 3), (2, 1, 1), (2,), (3, 2)],
        ]
    )
    def test_shapes_match_expected(
        self,
        permittivity_batch_shape,
        wavelength_batch_shape,
        excitation_batch_shape,
        num_unit_cells,
    ):
        layer_solve_results, thicknesses, s_matrices_interior = example_solve(
            permittivity_batch_shape, wavelength_batch_shape
        )
        num_terms = layer_solve_results[0].expansion.num_terms
        forward_amplitude_0_start = jnp.zeros(
            (2 * num_terms,) + excitation_batch_shape, dtype=complex
        )
        forward_amplitude_0_start = forward_amplitude_0_start.at[0, :].set(1)
        backward_amplitude_N_end = jnp.zeros_like(forward_amplitude_0_start)
        ab = fields.stack_amplitudes_interior(
            s_matrices_interior, forward_amplitude_0_start, backward_amplitude_N_end
        )
        ef, hf, (x, y, z) = fields.stack_fields_3d_auto_grid(
            amplitudes_interior=ab,
            layer_solve_results=layer_solve_results,
            layer_thicknesses=thicknesses,
            grid_spacing=0.2,
            num_unit_cells=num_unit_cells,
        )
        batch_shape = jnp.broadcast_shapes(
            permittivity_batch_shape, wavelength_batch_shape
        )
        expected_grid_shape = (5 * num_unit_cells[0], 5 * num_unit_cells[1])
        self.assertSequenceEqual(
            ef.shape,
            (3,) + batch_shape + expected_grid_shape + (200,) + excitation_batch_shape,
        )

        # Check that `z` is monotonic.
        self.assertTrue(not onp.any(onp.diff(z) < 0))

    def test_batch_excitation_matches_single(self):
        layer_solve_results, thicknesses, s_matrices_interior = example_solve((), ())
        num_terms = layer_solve_results[0].expansion.num_terms

        # Single amplitude solve.
        forward_amplitude_0_start = jnp.zeros((2 * num_terms, 1), dtype=complex)
        forward_amplitude_0_start = forward_amplitude_0_start.at[0, :].set(1)
        backward_amplitude_N_end = jnp.zeros_like(forward_amplitude_0_start)
        ab = fields.stack_amplitudes_interior(
            s_matrices_interior, forward_amplitude_0_start, backward_amplitude_N_end
        )
        ef_single, hf_single, (x, y, z) = fields.stack_fields_3d_auto_grid(
            amplitudes_interior=ab,
            layer_solve_results=layer_solve_results,
            layer_thicknesses=thicknesses,
            grid_spacing=0.2,
            num_unit_cells=(1, 1),
        )

        # Batch amplitude solve.
        forward_amplitude_0_start = jnp.zeros((2 * num_terms, 1), dtype=complex)
        forward_amplitude_0_start = forward_amplitude_0_start.at[0, :].set(1)
        backward_amplitude_N_end = jnp.zeros_like(forward_amplitude_0_start)
        ab = fields.stack_amplitudes_interior(
            s_matrices_interior, forward_amplitude_0_start, backward_amplitude_N_end
        )
        ef_batch, hf_batch, (x, y, z) = fields.stack_fields_3d_auto_grid(
            amplitudes_interior=ab,
            layer_solve_results=layer_solve_results,
            layer_thicknesses=thicknesses,
            grid_spacing=0.2,
            num_unit_cells=(1, 1),
        )

        onp.testing.assert_array_equal(ef_batch[..., 0], ef_single[..., 0])
        onp.testing.assert_array_equal(hf_batch[..., 0], hf_single[..., 0])
        onp.testing.assert_array_equal(ef_batch[..., 1], ef_single[..., 0])
        onp.testing.assert_array_equal(hf_batch[..., 1], hf_single[..., 0])

    def test_sequence_lengths_match_validation(self):
        with self.assertRaisesRegex(ValueError, "Encountered incompatible lengths"):
            fields.stack_fields_3d(
                amplitudes_interior=range(5),
                layer_solve_results=range(5),
                layer_thicknesses=range(5),
                layer_znum=range(4),
                grid_shape=(10, 10),
                num_unit_cells=(1, 1),
            )



"""Tests for `examples.metal_pillars`.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

import unittest

import jax.numpy as jnp
import numpy as onp

from examples import metal_pillars
from fmmax import fmm


class MetalPillarsTest(unittest.TestCase):
    def test_regression(self):
        # Checks that results match reference values. This helps protect against
        # regressions in accuracy of the simulator.
        n, rte, _, _ = metal_pillars.simulate_pillars(
            wavelength_nm=jnp.array([450.0]),
            approximate_num_terms=100,
            ambient_thickness_nm=0.0,
            formulation=fmm.Formulation.NORMAL,
        )
        onp.testing.assert_allclose(rte, [0.098347 + 0.099891j], rtol=1e-3)

    def test_compute_fields_regression(self):
        (
            _,
            _,
            _,
            (layer_solve_results, thicknesses, s_matrices_interior),
        ) = metal_pillars.simulate_pillars(
            wavelength_nm=jnp.array([450.0, 550.0]),
            approximate_num_terms=100,
            formulation=fmm.Formulation.NORMAL,
        )
        efields, hfields = metal_pillars.compute_fields(
            layer_solve_results=layer_solve_results,
            thicknesses=thicknesses,
            s_matrices_interior=s_matrices_interior,
            resolution_nm=5.0,
        )
        self.assertSequenceEqual(efields.shape, (3, 2, 36, 36, 240, 1))
        self.assertSequenceEqual(hfields.shape, (3, 2, 36, 36, 240, 1))

        with self.subTest("efields"):
            onp.testing.assert_allclose(
                jnp.mean(jnp.abs(efields) ** 2, axis=(2, 3, 4, 5)),
                onp.asarray(
                    [[0.046332, 0.053933], [1.048159, 1.887793], [0.145415, 0.080543]]
                ),
                rtol=6e-3,
            )
        with self.subTest("hfields"):
            onp.testing.assert_allclose(
                jnp.mean(jnp.abs(hfields) ** 2, axis=(2, 3, 4, 5)),
                onp.asarray(
                    [[1.038081, 1.833698], [0.004961, 0.003504], [0.045122, 0.050177]]
                ),
                rtol=6e-3,
            )



"""An example which simulates a micro-LED (uLED) with an internal source.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

import functools
import time
from typing import Sequence, Tuple

import jax.numpy as jnp

from fmmax import basis, fields, fmm, layer, scattering, sources


def simulate_uled(
    permittivity_ambient: complex = (1.0 + 0.0j),
    permittivity_passivation: complex = (1.5 + 0.0j) ** 2,
    permittivity_epi: complex = (3.0 + 0.0j) ** 2,
    permittivity_metal: complex = (0.2 + 3.3j) ** 2,
    thickness_ambient: float = 0.0,
    thickness_epi_top: float = 500.0,
    thickness_epi_bottom: float = 500.0,
    thickness_passivation: float = 100.0,
    thickness_metal: float = 100.0,
    pitch: float = 1400.0,
    epi_diameter: float = 1000.0,
    resolution: float = 1.0,
    resolution_fields: float = 10.0,
    wavelength: float = 620.0,
    dipole_y_offset: Sequence[float] = (0.0, 0.1, 0.2, 0.3, 0.4),
    dipole_fwhm: float = 20.0,
    approximate_num_terms: int = 1200,
    truncation: basis.Truncation = basis.Truncation.CIRCULAR,
    formulation: fmm.Formulation = fmm.Formulation.POL,
    brillouin_grid_shape: Tuple[int, int] = (1, 1),
) -> Tuple[
    jnp.ndarray,
    jnp.ndarray,
    jnp.ndarray,
    jnp.ndarray,
    Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray],
]:
    """Computes the extraction efficiency and fields for a uLED.

    The uLED consists of an epi (semiconductor) region, surrounded by a
    passivation dielectric and a metal. The structure is circularly
    symmetric, and the diameter of the epi region is `epi_diameter`.

    Dipoles are located in a plane partway through the epi region.
    Each dipole is located at `(x, y) = (0, offset)` for each of the
    offsets in `dipole_y_offset`. At each location, we model dipoles
    oriented in x-, y-, and z-directions, and each dipole has a spatial
    profile that is Gaussian with full-width at half-max given by
    `dipole_fwhm`.

    A cross section depiction of the structure is shown below.
                 _______________________________________
                |    |  |   epi                 |  |    |
                |    |  |                       |  |    |
                |    |  | - - - - - X - X - X - |  |    |
                |    |  |            dipoles    |  |    |
                |    |  |_______________________|  |    |
                |    |___passivation_______________|    |
                |___metal_______________________________|

    The default calculation models a periodic array of dipoles in a
    periodic array of uLEDs. To model an isolated dipole in a periodic
    array of uLEDs, a `brillouin_grid_shape` larger than the default
    `(1, 1)` can be specified, in which case the calculation is done
    for a grid of k-points in the first Brillouin zone. These can then
    be used to carry out Brillouin zone integration [2022 Lopez-Fraguas].

    [2022 Lopez-Fraguas] E. Lopez-Fraguas et al., "Tripling the light
        extraction efficiency of a deep ultraviolet LED using a
        nanostructured p-contact"
        https://www.nature.com/articles/s41598-022-15499-7

    Args:
        permittivity_ambient: The permittivity of the ambient.
        permittivity_passivation: The permittiity of the passivation.
        permittivity_epi: The permittiity of the epi material.
        permittivity_metal: The permittiity of the metal.
        thickness_ambient: The thickness of the ambient above the uLED.
        thickness_epi_top: The thickness of the epi layer above the source.
        thickness_epi_bottom: The thickness of the epi layer below the source.
        thickness_passivation: The thickness of the passivation layer.
        thickness_metal: The thickness of the metal below the passivation.
        pitch: The pitch of the uLED, i.e. the width.
        epi_diameter: The diameter of the epi region.
        resolution: The resolution of the real-space grid for permittivities.
        resolution_fields: The resolution of the real-space grid for fields.
        wavelength: The wavelength of the calculation.
        dipole_y_offset: Sequence of offsets for the dipole along the y axis.
            The dipole position is at `epi_diameter * dipole_y_offset`.
        dipole_fwhm: The dipole spatial full-width at half-maximum amplitude.
        approximate_num_terms: The approximate number of terms used in the
            plane wave expansion of the fields.
        truncation: Determines how the Fourier expansion is truncated.
        formulation: Specifies the formulation to be used.
        brillouin_grid_shape: The shape of the grid of k-points in the first
            Brillouin zone for which the calculation is done.

    Returns:
        extraction_efficiency: The computed extraction efficiency.
        total_emitted_power: The total power emitted by the dipole.
        efields: The electric fields in the structure, with shape `(3,)
            + brillouin_grid_shape + (xnum, ynum, znum, num_dipoles)`.
        hfields: The magnetic fields with shape matching `efields`.
        grid: The coordinates `(x, y, z)` for the fields.
    """
    (
        permittivities_before_source,
        permittivities_after_source,
        thicknesses_before_source,
        thicknesses_after_source,
        primitive_lattice_vectors,
    ) = uled_structure(
        permittivity_ambient=permittivity_ambient,
        permittivity_passivation=permittivity_passivation,
        permittivity_epi=permittivity_epi,
        permittivity_metal=permittivity_metal,
        thickness_ambient=thickness_ambient,
        thickness_epi_top=thickness_epi_top,
        thickness_epi_bottom=thickness_epi_bottom,
        thickness_passivation=thickness_passivation,
        thickness_metal=thickness_metal,
        pitch=pitch,
        epi_diameter=epi_diameter,
        resolution=resolution,
    )

    expansion = basis.generate_expansion(
        primitive_lattice_vectors=primitive_lattice_vectors,
        approximate_num_terms=approximate_num_terms,
        truncation=truncation,
    )

    in_plane_wavevector = basis.brillouin_zone_in_plane_wavevector(
        brillouin_grid_shape, primitive_lattice_vectors
    )
    assert in_plane_wavevector.shape[-1] == 2
    assert in_plane_wavevector.ndim == 3

    eigensolve = functools.partial(
        layer.eigensolve_isotropic_media,
        wavelength=jnp.asarray(wavelength),
        in_plane_wavevector=in_plane_wavevector,
        primitive_lattice_vectors=primitive_lattice_vectors,
        expansion=expansion,
        formulation=formulation,
    )

    # Perform the layer eigensolve for each layer in the stack. Use the fact that
    # the first layer after the source is identical to the layer just before the
    # source in order to avoid one eigensolve.
    layer_solve_results_before_source = tuple(
        [eigensolve(permittivity=p) for p in permittivities_before_source]
    )
    layer_solve_results_after_source = tuple(
        [layer_solve_results_before_source[-1]]
        + [eigensolve(permittivity=p) for p in permittivities_after_source[1:]]
    )

    # Compute interior scattering matrices.
    s_matrices_interior_before_source = scattering.stack_s_matrices_interior(
        layer_solve_results_before_source, thicknesses_before_source
    )
    s_matrices_interior_after_source = scattering.stack_s_matrices_interior(
        layer_solve_results_after_source, thicknesses_after_source
    )
    s_matrix_before_source = s_matrices_interior_before_source[-1][0]
    s_matrix_after_source = s_matrices_interior_after_source[-1][0]

    # TODO: update the locations and remove the shift by `resolution`. Requires
    # updating regression test values.
    dipole_locations = [
        (pitch / 2 - resolution, pitch / 2 + offset * epi_diameter - resolution)
        for offset in dipole_y_offset
    ]
    dipoles = sources.gaussian_source(
        fwhm=jnp.asarray(dipole_fwhm),
        location=jnp.asarray(dipole_locations),
        in_plane_wavevector=in_plane_wavevector,
        primitive_lattice_vectors=primitive_lattice_vectors,
        expansion=expansion,
    )

    zeros = jnp.zeros_like(dipoles)
    jx = jnp.concatenate([dipoles, zeros, zeros], axis=-1)
    jy = jnp.concatenate([zeros, dipoles, zeros], axis=-1)
    jz = jnp.concatenate([zeros, zeros, dipoles], axis=-1)

    (
        bwd_amplitude_ambient_end,
        fwd_amplitude_before_start,
        bwd_amplitude_before_end,
        fwd_amplitude_after_start,
        bwd_amplitude_after_end,
        fwd_amplitude_substrate_start,
    ) = sources.amplitudes_for_source(
        jx=jx,
        jy=jy,
        jz=jz,
        s_matrix_before_source=s_matrix_before_source,
        s_matrix_after_source=s_matrix_after_source,
    )

    # Compute the Poynting flux just before the source, after the source,
    # and in the ambient, to allow calculation of the extraction efficiency.
    fwd_amplitude_before_end = fields.propagate_amplitude(
        amplitude=fwd_amplitude_before_start,
        distance=s_matrix_before_source.end_layer_thickness,
        layer_solve_result=s_matrix_before_source.end_layer_solve_result,
    )
    fwd_flux_before_end, bwd_flux_before_end = fields.directional_poynting_flux(
        forward_amplitude=fwd_amplitude_before_end,
        backward_amplitude=bwd_amplitude_before_end,
        layer_solve_result=s_matrix_before_source.end_layer_solve_result,
    )
    bwd_amplitude_after_start = fields.propagate_amplitude(
        amplitude=bwd_amplitude_after_end,
        distance=s_matrix_after_source.start_layer_thickness,
        layer_solve_result=s_matrix_after_source.start_layer_solve_result,
    )
    fwd_flux_after_start, bwd_flux_after_start = fields.directional_poynting_flux(
        forward_amplitude=fwd_amplitude_after_start,
        backward_amplitude=bwd_amplitude_after_start,
        layer_solve_result=s_matrix_after_source.start_layer_solve_result,
    )
    _, bwd_flux_ambient_end = fields.directional_poynting_flux(
        forward_amplitude=jnp.zeros_like(bwd_amplitude_ambient_end),
        backward_amplitude=bwd_amplitude_ambient_end,
        layer_solve_result=s_matrix_before_source.start_layer_solve_result,
    )

    # Compute the total forward and backward flux resulting from the source. The
    # forward flux from the source is the difference between the forward flux just
    # after the source, and the forward flux just before the source. The backward
    # flux is defined analogously.
    fwd_flux_from_source = fwd_flux_after_start - fwd_flux_before_end
    bwd_flux_from_source = bwd_flux_before_end - bwd_flux_after_start

    # Compute total power emitted by the source by summing over the Brillouin
    # zone points and Fourier orders.
    assert fwd_flux_from_source.shape == brillouin_grid_shape + (
        2 * expansion.num_terms,
        jx.shape[-1],
    )
    forward_emitted_power = jnp.sum(fwd_flux_from_source, axis=(-4, -3, -2))
    backward_emitted_power = -jnp.sum(bwd_flux_from_source, axis=(-4, -3, -2))
    total_emitted_power = forward_emitted_power + backward_emitted_power

    total_extracted_power = -jnp.sum(bwd_flux_ambient_end, axis=(-4, -3, -2))
    assert total_extracted_power.shape == (jx.shape[-1],)

    # Calculate the per-dipole extraction efficiency.
    extraction_efficiency = total_extracted_power / total_emitted_power

    # Compute the fields inside the structure.
    amplitudes_interior = fields.stack_amplitudes_interior_with_source(
        s_matrices_interior_before_source=s_matrices_interior_before_source,
        s_matrices_interior_after_source=s_matrices_interior_after_source,
        backward_amplitude_before_end=bwd_amplitude_before_end,
        forward_amplitude_after_start=fwd_amplitude_after_start,
    )
    efields, hfields, (x, y, z) = fields.stack_fields_3d_auto_grid(
        amplitudes_interior=amplitudes_interior,
        layer_solve_results=(
            layer_solve_results_before_source + layer_solve_results_after_source
        ),
        layer_thicknesses=thicknesses_before_source + thicknesses_after_source,
        grid_spacing=resolution_fields,
        num_unit_cells=brillouin_grid_shape,
    )

    return extraction_efficiency, total_emitted_power, efields, hfields, (x, y, z)


def uled_structure(
    permittivity_ambient: complex,
    permittivity_passivation: complex,
    permittivity_epi: complex,
    permittivity_metal: complex,
    thickness_ambient: float,
    thickness_epi_top: float,
    thickness_epi_bottom: float,
    thickness_passivation: float,
    thickness_metal: float,
    pitch: float,
    epi_diameter: float,
    resolution: float,
) -> Tuple[
    Tuple[jnp.ndarray, ...],
    Tuple[jnp.ndarray, ...],
    Tuple[jnp.ndarray, ...],
    Tuple[jnp.ndarray, ...],
    basis.LatticeVectors,
]:
    """Returns quantities that describe the uLED structure."""

    primitive_lattice_vectors = basis.LatticeVectors(
        u=basis.X * pitch, v=basis.Y * pitch
    )

    # Generate the permittivity for the cross section below the epi
    # structure, including only the metal and the passivation.
    outer_circle_diameter = epi_diameter + thickness_passivation * 2
    inside_outer_circle = circle_mask(
        pitch, outer_circle_diameter, resolution, x_offset=0, y_offset=0
    )
    permittivity_cross_section_passivation_only = jnp.where(
        inside_outer_circle, permittivity_passivation, permittivity_metal
    )

    # Permittivity for the cross section also including the epi.
    inside_inner_circle = circle_mask(
        pitch, epi_diameter, resolution, x_offset=0, y_offset=0
    )
    permittivity_cross_section_epi = jnp.where(
        inside_inner_circle,
        permittivity_epi,
        permittivity_cross_section_passivation_only,
    )

    # Generate the list of permittivities. For those which are scalars,
    # we add singleton spatial dimensions. Note that the epi region is
    # split into two layers, between which the source will be inserted.
    permittivities_before_source = [
        jnp.asarray(permittivity_ambient)[jnp.newaxis, jnp.newaxis],
        permittivity_cross_section_epi,
    ]
    permittivities_after_source = [
        permittivity_cross_section_epi,
        permittivity_cross_section_passivation_only,
        jnp.asarray(permittivity_metal)[jnp.newaxis, jnp.newaxis],
    ]

    thicknesses_before_source = [
        jnp.asarray(thickness_ambient),
        jnp.asarray(thickness_epi_top),
    ]
    thicknesses_after_source = [
        jnp.asarray(thickness_epi_bottom),
        jnp.asarray(thickness_passivation),
        jnp.asarray(thickness_metal),
    ]

    return (
        tuple(permittivities_before_source),
        tuple(permittivities_after_source),
        tuple(thicknesses_before_source),
        tuple(thicknesses_after_source),
        primitive_lattice_vectors,
    )


def circle_mask(
    pitch: float,
    diameter: float,
    resolution: float,
    x_offset: float,
    y_offset: float,
) -> jnp.ndarray:
    """Returns a mask that is `True` for a centered circular feature."""
    x, y = jnp.meshgrid(
        jnp.arange(-pitch / 2 + resolution / 2, pitch / 2, resolution),
        jnp.arange(-pitch / 2 + resolution / 2, pitch / 2, resolution),
        indexing="ij",
    )
    distance = jnp.sqrt((x - x_offset) ** 2 + (y - y_offset) ** 2)
    return distance < diameter / 2


if __name__ == "__main__":
    for approximate_num_terms in (200, 400, 600, 800, 1000, 1200, 1500, 2000, 2500):
        t0 = time.time()
        extraction_efficiency, total_emitted_power, *_ = simulate_uled(
            approximate_num_terms=approximate_num_terms,
        )
        print(
            f"Results with approximate_num_terms = {approximate_num_terms} (time = {time.time() - t0:.2f}s):\n"
            f"    Extraction efficiency = {jnp.squeeze(extraction_efficiency * 100)}%\n"
            f"      Total emitted power = {jnp.squeeze(total_emitted_power)}"
        )



"""Tests for `examples.ar_coating`.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

import unittest

import jax.numpy as jnp
import numpy as onp

from examples import ar_coating


class ARCoatingTest(unittest.TestCase):
    def test_refractive_index_shape_validation(self):
        with self.assertRaisesRegex(ValueError, "All refractive indices"):
            ar_coating.compute_reflection(
                refractive_indices=(jnp.asarray([[2.0]]),),
                thicknesses=(100.0),
                refractive_index_ambient=1.0,
                refractive_index_substrate=1.45,
                wavelength=jnp.asarray(450.0),
                incident_angle=jnp.asarray(0.0),
            )

    def test_compare_to_lumerical(self):
        wavelength = jnp.asarray([450.0, 520.0, 640.0])
        incident_angle = jnp.asarray(
            [0.0, 15 * jnp.pi / 180, 30 * jnp.pi / 180, 45 * jnp.pi / 180]
        )

        rte, rtm = ar_coating.compute_reflection(
            refractive_indices=(1.0, 1.45, 2.0, 1.45, 2.0, 1.45),
            thicknesses=(1000.0, 92.0, 75.0, 24.0, 28.0, 1000.0),
            refractive_index_ambient=1.0,
            refractive_index_substrate=1.45,
            # The second batch dimension is for the wavelength.
            wavelength=wavelength[jnp.newaxis, :],
            # The first batch dimension is for the incident angle.
            incident_angle=incident_angle[:, jnp.newaxis],
        )

        # Reference TE reflection coefficient values computed by Lumerical stack.
        lumerical_efield_rte = jnp.asarray(
            [
                [-0.0116697 - 0.029851j, -0.0403836 + 0.050517j, 0.0210578 - 0.037780j],
                [-0.0164556 - 0.0145856j, 0.0100469 + 0.07167j, -0.0090968 - 0.052744j],
                [-0.0365618 + 0.0335138j, 0.054515 - 0.07834j, -0.0720199 + 0.0533403j],
                [-0.0505926 - 0.12564j, 0.00661052 + 0.139193j, 0.116759 - 0.119862j],
            ]
        )
        expected_rte = -lumerical_efield_rte
        onp.testing.assert_allclose(
            jnp.abs(rte) ** 2, jnp.abs(expected_rte) ** 2, rtol=1e-4
        )
        onp.testing.assert_allclose(rte, expected_rte, rtol=1e-4)

        # Reference TM reflection coefficient values computed by Lumerical stack.
        lumerical_efield_rtm = jnp.asarray(
            [
                [-0.0116697 - 0.029851j, -0.0403836 + 0.050517j, 0.0210578 - 0.037780j],
                [-0.0178055 - 0.005578j, 0.0080347 + 0.059176j, 0.0009481 - 0.0489879j],
                [-0.0072645 - 0.003216j, 0.0235415 - 0.029488j, -0.070394 + 0.0028323j],
                [0.0306234 + 0.0026343j, -0.014264 - 0.0177914j, 0.103417 + 0.0322221j],
            ]
        )
        expected_rtm = -lumerical_efield_rtm
        onp.testing.assert_allclose(
            jnp.abs(rtm) ** 2, jnp.abs(expected_rtm) ** 2, rtol=1e-4
        )
        onp.testing.assert_allclose(rtm, expected_rtm, rtol=1e-4)



"""Tests for `examples.sorter`.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

import unittest

import numpy as onp

from examples import sorter


class SorterTest(unittest.TestCase):
    def test_regression(self):
        values = sorter.optimize(steps=2, approximate_num_terms=30)
        expected_values = [3.257598, 3.249871]
        onp.testing.assert_allclose(values, expected_values, rtol=1e-4)



"""Tests for `fmmax.vector`.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

import unittest

import jax
import jax.numpy as jnp
import numpy as onp
import parameterized

from fmmax import basis, vector


class ChangeBasisTest(unittest.TestCase):
    def test_basis_cycle(self):
        x = jnp.array([0.2, 0.8])
        y = jnp.array([-0.1, 0.3])
        u = jnp.array([-0.1, -0.7])
        v = jnp.array([0.0, 1.2])
        tu = jax.random.uniform(jax.random.PRNGKey(0), (10, 12))
        tv = jax.random.uniform(jax.random.PRNGKey(1), (10, 12))
        tx, ty = vector.change_vector_field_basis(tu, tv, u, v, x, y)
        tu_recovered, tv_recovered = vector.change_vector_field_basis(
            tx, ty, x, y, u, v
        )
        onp.testing.assert_allclose(tu_recovered, tu, rtol=1e-4)
        onp.testing.assert_allclose(tv_recovered, tv, rtol=1e-4)


class NormalizeTest(unittest.TestCase):
    @parameterized.parameterized.expand(
        [
            (
                vector.normalize_normal,
                [[1 / jnp.sqrt(2), 1.0, 0.0, 1.0, 0.0]],
                [[1 / jnp.sqrt(2), 0.0, 1.0, 0.0, 0.0]],
            ),
            (
                vector.normalize_pol,
                [[1 / jnp.sqrt(2), 0.2 / jnp.sqrt(2), 0.0, 0.01 / jnp.sqrt(2), 0.0]],
                [[1 / jnp.sqrt(2), 0.0, 0.2 / jnp.sqrt(2), 0.0, 0.0]],
            ),
            (
                vector.normalize_jones,
                [[0.5 + 0.5j, 0.734, 0.680, 0.707, 0.707]],
                [[0.5 + 0.5j, 0.680j, 0.734j, 0.707j, 0.707j]],
            ),
        ]
    )
    def test_normalized_matches_expected(self, normalize_fn, expected_tx, expected_ty):
        tx = jnp.asarray([[1.0, 0.2, 0.0, 0.01, 0.0]], dtype=float)
        ty = jnp.asarray([[1.0, 0.0, 0.2, 0.0, 0.0]], dtype=float)
        tx, ty = normalize_fn(tx, ty)
        onp.testing.assert_allclose(tx, jnp.asarray(expected_tx), rtol=1e-3)
        onp.testing.assert_allclose(ty, jnp.asarray(expected_ty), rtol=1e-3)

    @parameterized.parameterized.expand(
        [(vector.normalize_normal,), (vector.normalize_pol,), (vector.normalize_jones,)]
    )
    def test_zeros_no_nan(self, normalize_fn):
        tx = jnp.zeros((20, 20))
        ty = jnp.zeros((20, 20))
        tx, ty = normalize_fn(tx, ty)
        self.assertFalse(onp.any(onp.isnan(tx)))
        self.assertFalse(onp.any(onp.isnan(ty)))

    @parameterized.parameterized.expand(
        [(vector.normalize_normal,), (vector.normalize_pol,), (vector.normalize_jones,)]
    )
    def test_gradient_no_nan(self, normalize_fn):
        def loss_fn(tx, ty):
            tx, ty = normalize_fn(tx, ty)
            return jnp.real(jnp.sum(tx) + jnp.sum(ty))

        gx, gy = jax.grad(loss_fn, argnums=(0, 1))(jnp.zeros((5, 5)), jnp.zeros((5, 5)))
        self.assertFalse(onp.any(onp.isnan(gx)))
        self.assertFalse(onp.any(onp.isnan(gy)))


# -----------------------------------------------------------------------------
# Tests related to the `tangent_field` function.
# -----------------------------------------------------------------------------


class TangentFieldTest(unittest.TestCase):
    def test_optimize(self):
        arr = jnp.array(
            [[0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0]], dtype=jnp.float32
        )
        tx, ty = vector.tangent_field(
            arr,
            use_jones=False,
            optimizer=vector.OPTIMIZER,
            alignment_weight=vector.ALIGNMENT_WEIGHT,
            smoothness_weight=vector.SMOOTHNESS_WEIGHT,
            steps_dim_multiple=vector.STEPS_DIM_MULTIPLE,
            smoothing_kernel=jnp.ones((1, 1)),
        )
        expected_tx = [
            [
                0.083,
                0.25,
                0.417,
                0.583,
                0.75,
                0.667,
                0.333,
                0.0,
                -0.333,
                -0.667,
                -0.75,
                -0.583,
                -0.417,
                -0.25,
                -0.083,
            ]
        ]
        onp.testing.assert_allclose(tx, expected_tx, rtol=0.02)
        onp.testing.assert_allclose(ty, 0.0, atol=1e-7)

    def test_optimize_jones(self):
        arr = jnp.array(
            [[0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0]], dtype=jnp.float32
        )
        tx, ty = vector.tangent_field(
            arr,
            use_jones=True,
            optimizer=vector.OPTIMIZER,
            alignment_weight=vector.ALIGNMENT_WEIGHT,
            smoothness_weight=vector.SMOOTHNESS_WEIGHT,
            steps_dim_multiple=vector.STEPS_DIM_MULTIPLE,
            smoothing_kernel=jnp.ones((1, 1)),
        )
        expected_tx_magnitude = jnp.ones_like(arr)
        onp.testing.assert_allclose(jnp.abs(tx), expected_tx_magnitude)
        onp.testing.assert_allclose(ty, 0.0, atol=1e-7)


class SchemesTest(unittest.TestCase):
    @parameterized.parameterized.expand(
        [(vector.JONES_DIRECT,), (vector.JONES,), (vector.POL,), (vector.NORMAL,)]
    )
    def test_batch_matches_single_exact(self, scheme):
        key = jax.random.PRNGKey(0)
        arr = jax.random.uniform(key, shape=(5, 10, 10))
        tx, ty = vector.VECTOR_FIELD_SCHEMES[scheme](
            arr=arr,
            primitive_lattice_vectors=basis.LatticeVectors(u=basis.X, v=basis.Y),
        )
        for i in range(5):
            expected_tx_i, expected_ty_i = vector.VECTOR_FIELD_SCHEMES[scheme](
                arr=arr[i, :, :],
                primitive_lattice_vectors=basis.LatticeVectors(u=basis.X, v=basis.Y),
            )
            onp.testing.assert_array_equal(tx[i, :, :], expected_tx_i)
            onp.testing.assert_array_equal(ty[i, :, :], expected_ty_i)

    @parameterized.parameterized.expand(
        [(scheme,) for scheme in vector.VECTOR_FIELD_SCHEMES]
    )
    def test_uniform_array_no_nan(self, scheme):
        # The tangent field calculation requires special logic to handle uniform arrays,
        #  otherwise `nan` will show up in the result.
        arr = jnp.ones((10, 10))
        tx, ty = vector.VECTOR_FIELD_SCHEMES[scheme](
            arr=arr,
            primitive_lattice_vectors=basis.LatticeVectors(u=basis.X, v=basis.Y),
        )
        self.assertFalse(onp.any(onp.isnan(tx)))
        self.assertFalse(onp.any(onp.isnan(ty)))

    @parameterized.parameterized.expand(
        [(scheme,) for scheme in vector.VECTOR_FIELD_SCHEMES]
    )
    def test_gradient_no_nan(self, scheme):
        def _loss_fn(arr):
            tx, ty = vector.VECTOR_FIELD_SCHEMES[scheme](
                arr=arr,
                primitive_lattice_vectors=basis.LatticeVectors(u=basis.X, v=basis.Y),
            )
            return jnp.sum(jnp.abs(tx) ** 2 + jnp.abs(ty) ** 2)

        arr = jnp.asarray(
            [
                [1.0 + 1.0j, 1.0, 1.0, 1.0, 1.0 + 1.1j, 1.0 + 1.1j],
                [1.0 + 1.0j, 0.0, 0.0, 0.0, 1.0 + 1.1j, 1.0 + 1.1j],
                [1.0 + 1.0j, 0.0, 0.0, 0.0, 1.0 + 1.1j, 1.0 + 1.1j],
                [1.0 + 1.0j, 1.0, 1.0, 1.0, 1.0 + 1.1j, 1.0 + 1.1j],
            ]
        )
        grad = jax.grad(_loss_fn)(arr)
        self.assertFalse(onp.any(onp.isnan(grad)))


class LossTest(unittest.TestCase):
    @parameterized.parameterized.expand(
        [
            (1.0, 0.0, 1.0, 0.0, -1.0),
            (-1.0, 0.0, 1.0, 0.0, -1.0),
            (0.0, 1.0, 1.0, 0.0, 0.0),
        ]
    )
    def test_self_alignment_loss(self, tx, ty, tx0, ty0, expected):
        tx = jnp.asarray(tx)[jnp.newaxis, jnp.newaxis]
        ty = jnp.asarray(ty)[jnp.newaxis, jnp.newaxis]
        tx0 = jnp.asarray(tx0)[jnp.newaxis, jnp.newaxis]
        ty0 = jnp.asarray(ty0)[jnp.newaxis, jnp.newaxis]
        loss = vector._self_alignment_loss(tx, ty, tx0, ty0)
        onp.testing.assert_allclose(loss, expected)

    @parameterized.parameterized.expand(
        [
            (
                jnp.ones((2, 2)),
                jnp.zeros((2, 2)),
                jnp.ones((2, 2)),
                jnp.zeros((2, 2)),
                -400,
            ),
            (
                jnp.asarray([[1, 1], [-1, -1], [-1, -1], [1, 1]]),
                jnp.zeros((4, 2)),
                jnp.ones((4, 2)),
                jnp.zeros((4, 2)),
                -768.0,
            ),
        ]
    )
    def test_field_loss(self, tx, ty, tx0, ty0, expected):
        loss = vector._field_loss(
            tx, ty, tx0, ty0, alignment_weight=100, smoothness_weight=2
        )
        onp.testing.assert_allclose(loss, expected)

    def test_field_loss_batch_matches_single(self):
        key = jax.random.PRNGKey(0)
        tx, ty, tx0, ty0 = jax.random.uniform(key, (4, 8, 5, 10))
        loss = vector._field_loss(
            tx, ty, tx0, ty0, alignment_weight=100, smoothness_weight=2
        )
        expected_loss = 0
        for tx_slice, ty_slice, tx0_slice, ty0_slice in zip(tx, ty, tx0, ty0):
            expected_loss += vector._field_loss(
                tx_slice,
                ty_slice,
                tx0_slice,
                ty0_slice,
                alignment_weight=100,
                smoothness_weight=2,
            )
        onp.testing.assert_allclose(loss, expected_loss, rtol=1e-6)


class UtilitiesTest(unittest.TestCase):
    @parameterized.parameterized.expand(
        [
            ((1.0, 1.0), 1.0, (jnp.sqrt(0.5), jnp.sqrt(0.5))),
            ((0.1, 0.1), 1.0, (0.1, 0.1)),
            ((2.0, 0.0), 1.0, (1.0, 0.0)),
            ((2.0, 0.0), 1.5, (1.5, 0.0)),
            ((2.0j, 0.0), 1.5, (1.5j, 0.0)),
        ]
    )
    def test_clip_magnitude(self, tx_ty, max_magnitude, expected):
        result = vector._clip_magnitude(*tx_ty, max_magnitude)
        onp.testing.assert_allclose(result, expected)

    @parameterized.parameterized.expand(
        [
            ((1.0, 0.0), (1.0, 0.0)),
            ((1.0j, 0.0), (1.0, 0.0)),
            ((1.0j, 1.0), (jnp.sqrt(0.5) * (1.0 + 1.0j), jnp.sqrt(0.5) * (1.0 - 1.0j))),
        ]
    )
    def test_remove_average_phase(self, tx_ty, expected):
        tx, ty = tx_ty
        tx = jnp.asarray(tx)[jnp.newaxis, jnp.newaxis]
        ty = jnp.asarray(ty)[jnp.newaxis, jnp.newaxis]
        result_tx, result_ty = vector._remove_mean_phase(tx, ty)
        onp.testing.assert_allclose(
            (result_tx.squeeze(), result_ty.squeeze()), expected
        )



"""Tests for `fmmax.basis`.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

import unittest

import jax
import jax.numpy as jnp
import numpy as onp
import parameterized

from fmmax import basis


def _coeffs_set(coeffs):
    return {tuple([int(idx) for idx in c]) for c in coeffs}


class ExpansionTest(unittest.TestCase):
    def test_circular_with_square_lattice(self):
        primitive_lattice_vectors = basis.LatticeVectors(
            u=jnp.array([1.0, 0.0]), v=jnp.array([0.0, 1.0])
        )
        expansion = basis.generate_expansion(
            primitive_lattice_vectors=primitive_lattice_vectors,
            approximate_num_terms=20,
            truncation=basis.Truncation.CIRCULAR,
        )
        onp.testing.assert_allclose(
            primitive_lattice_vectors.reciprocal.u, onp.array([1.0, 0.0])
        )
        onp.testing.assert_allclose(
            primitive_lattice_vectors.reciprocal.v, onp.array([0.0, 1.0])
        )
        # Circular truncation discards the corner elements.
        circular_mask = jnp.array(
            [
                [0, 1, 1, 1, 0],
                [1, 1, 1, 1, 1],
                [1, 1, 1, 1, 1],
                [1, 1, 1, 1, 1],
                [0, 1, 1, 1, 0],
            ]
        ).astype(bool)
        i, j = jnp.meshgrid(jnp.arange(-2, 3), jnp.arange(-2, 3), indexing="ij")
        expected_coefficients = jnp.stack(
            [i[circular_mask].flatten(), j[circular_mask].flatten()], axis=-1
        )
        self.assertSequenceEqual(
            _coeffs_set(expansion.basis_coefficients),
            _coeffs_set(expected_coefficients),
        )

    def test_circular_with_rectangle_lattice(self):
        primitive_lattice_vectors = basis.LatticeVectors(
            u=jnp.array([2.0, 0.0]), v=jnp.array([0.0, 1.0])
        )
        expansion = basis.generate_expansion(
            primitive_lattice_vectors=primitive_lattice_vectors,
            approximate_num_terms=20,
            truncation=basis.Truncation.CIRCULAR,
        )
        onp.testing.assert_allclose(
            primitive_lattice_vectors.reciprocal.u, onp.array([0.5, 0.0])
        )
        onp.testing.assert_allclose(
            primitive_lattice_vectors.reciprocal.v, onp.array([0.0, 1.0])
        )
        # Circular truncation discards the corner elements.
        circular_mask = jnp.array(
            [
                [0, 1, 0],
                [1, 1, 1],
                [1, 1, 1],
                [1, 1, 1],
                [1, 1, 1],
                [1, 1, 1],
                [0, 1, 0],
            ]
        ).astype(bool)
        i, j = jnp.meshgrid(jnp.arange(-3, 4), jnp.arange(-1, 2), indexing="ij")
        expected_coefficients = jnp.stack(
            [i[circular_mask].flatten(), j[circular_mask].flatten()], axis=-1
        )
        self.assertSequenceEqual(
            _coeffs_set(expansion.basis_coefficients),
            _coeffs_set(expected_coefficients),
        )

    def test_parallelogramic_with_square_lattice(self):
        primitive_lattice_vectors = basis.LatticeVectors(
            u=jnp.array([1.0, 0.0]), v=jnp.array([0.0, 1.0])
        )
        expansion = basis.generate_expansion(
            primitive_lattice_vectors=primitive_lattice_vectors,
            approximate_num_terms=20,
            truncation=basis.Truncation.PARALLELOGRAMIC,
        )
        onp.testing.assert_allclose(
            primitive_lattice_vectors.reciprocal.u, onp.array([1.0, 0.0])
        )
        onp.testing.assert_allclose(
            primitive_lattice_vectors.reciprocal.v, onp.array([0.0, 1.0])
        )
        # Parallelogramic truncation includes all coefficients in the range `(-2, +2)`.
        i, j = jnp.meshgrid(jnp.arange(-2, 3), jnp.arange(-2, 3), indexing="ij")
        expected_coefficients = jnp.stack([i.flatten(), j.flatten()], axis=-1)
        self.assertSequenceEqual(
            _coeffs_set(expansion.basis_coefficients),
            _coeffs_set(expected_coefficients),
        )

    def test_parallelogramic_with_rectangle_lattice(self):
        primitive_lattice_vectors = basis.LatticeVectors(
            u=jnp.array([2.0, 0.0]), v=jnp.array([0.0, 1.0])
        )
        expansion = basis.generate_expansion(
            primitive_lattice_vectors=primitive_lattice_vectors,
            approximate_num_terms=20,
            truncation=basis.Truncation.PARALLELOGRAMIC,
        )
        onp.testing.assert_allclose(
            primitive_lattice_vectors.reciprocal.u, onp.array([0.5, 0.0])
        )
        onp.testing.assert_allclose(
            primitive_lattice_vectors.reciprocal.v, onp.array([0.0, 1.0])
        )
        i, j = jnp.meshgrid(jnp.arange(-3, 4), jnp.arange(-1, 2), indexing="ij")
        expected_coefficients = jnp.stack([i.flatten(), j.flatten()], axis=-1)
        self.assertSequenceEqual(
            _coeffs_set(expansion.basis_coefficients),
            _coeffs_set(expected_coefficients),
        )

    def test_expansion_treedef(self):
        # Checks that equality checks for the treedef of an `Expansion` are possible.
        primitive_lattice_vectors = basis.LatticeVectors(
            u=jnp.array([2.0, 0.0]), v=jnp.array([0.0, 1.0])
        )
        expansion = basis.generate_expansion(
            primitive_lattice_vectors=primitive_lattice_vectors,
            approximate_num_terms=20,
            truncation=basis.Truncation.PARALLELOGRAMIC,
        )
        treedef = jax.tree_util.tree_structure(expansion)
        self.assertEqual(treedef, treedef)

    def test_expansion_flatten_unflatten(self):
        # Checks that equality checks for the treedef of an `Expansion` are possible.
        primitive_lattice_vectors = basis.LatticeVectors(
            u=jnp.array([2.0, 0.0]), v=jnp.array([0.0, 1.0])
        )
        expansion = basis.generate_expansion(
            primitive_lattice_vectors=primitive_lattice_vectors,
            approximate_num_terms=20,
            truncation=basis.Truncation.PARALLELOGRAMIC,
        )
        leaves, treedef = jax.tree_util.tree_flatten(expansion)
        restored = jax.tree_util.tree_unflatten(treedef, leaves)


class InPlaneWavevectorTest(unittest.TestCase):
    @parameterized.parameterized.expand([[(1,)], [(1, 2, 3)], [(0, 1)]])
    def test_brillouin_grid_shape_validation(self, invalid_shape):
        with self.assertRaisesRegex(
            ValueError, "`brillouin_grid_shape` must be length-2 with"
        ):
            basis.brillouin_zone_in_plane_wavevector(
                brillouin_grid_shape=invalid_shape,
                primitive_lattice_vectors=basis.LatticeVectors(basis.X, basis.Y),
            )

    @parameterized.parameterized.expand(
        [
            [(1, 1), [[[0, 0]]]],
            [(2, 1), [[[-0.25 * jnp.pi, 0]], [[0.25 * jnp.pi, 0]]]],
            [(1, 2), [[[0, -0.5 * jnp.pi], [0, 0.5 * jnp.pi]]]],
        ],
    )
    def test_brillouin_wavevector_matches_exected(self, shape, expected_vectors):
        primitive_lattice_vectors = basis.LatticeVectors(basis.X * 2, basis.Y)
        wavevector = basis.brillouin_zone_in_plane_wavevector(
            brillouin_grid_shape=shape,
            primitive_lattice_vectors=primitive_lattice_vectors,
        )
        onp.testing.assert_allclose(wavevector, expected_vectors)

    @parameterized.parameterized.expand(
        [
            [1.0, 0.0, 0.0, 1.0, (0, 0)],
            [1.0, jnp.pi / 4, 0.0, 1.0, (2 * jnp.pi / jnp.sqrt(2), 0)],
            [1.0, jnp.pi / 4, jnp.pi / 2, 1.0, (0, 2 * jnp.pi / jnp.sqrt(2))],
        ]
    )
    def test_plane_wave_wavevector_matches_expected(
        self, wavelength, polar_angle, azimuthal_angle, permittivity, expected
    ):
        wavevector = basis.plane_wave_in_plane_wavevector(
            wavelength=wavelength,
            polar_angle=polar_angle,
            azimuthal_angle=azimuthal_angle,
            permittivity=permittivity,
        )
        onp.testing.assert_allclose(wavevector, expected, rtol=1e-6, atol=1e-6)


class UnitCellCoordiantesTest(unittest.TestCase):
    @parameterized.parameterized.expand(
        [
            [(1, 1), (3, 2), (0, 1, 2), (0, 1)],
            [(1, 1), (4, 3), (0, 1, 2, 3), (0, 1, 2)],
            [(2, 2), (2, 1), (0, 0.5, 1.0, 1.5), (0, 0.5)],
        ]
    )
    def test_values_match_expected(self, shape, num_unit_cells, expected_x, expected_y):
        x, y = basis.unit_cell_coordinates(
            primitive_lattice_vectors=basis.LatticeVectors(u=basis.X, v=basis.Y),
            shape=shape,
            num_unit_cells=num_unit_cells,
        )
        onp.testing.assert_array_equal(
            x, jnp.broadcast_to(jnp.asarray(expected_x)[:, jnp.newaxis], x.shape)
        )
        onp.testing.assert_array_equal(
            y, jnp.broadcast_to(jnp.asarray(expected_y)[jnp.newaxis, :], y.shape)
        )



"""Functions related to tangent vector field generation.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

import functools
from typing import Any, Callable, Dict, Tuple

import jax
import jax.example_libraries.optimizers as jopt
import jax.numpy as jnp
import numpy as onp

from fmmax import basis, utils

PyTree = Any


def normalized_vector_field(
    arr: jnp.ndarray,
    primitive_lattice_vectors: basis.LatticeVectors,
    vector_fn: Callable,
    normalize_fn: Callable,
    resize_max_dim: int,
    resize_method: jax.image.ResizeMethod,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Generates a normalized tangent vector field according to the specified method.

    Some `vector_fn` can be computationally expensive, and so this function
    resizes the input `arr` so that the maximum size of the two trailing
    dimensions is `resize_max_dim`. When `arr` is smaller than the maximum
    size, no resampling is performed.

    The tangent fields are then computed for this resized array, and then
    resized again to obtain fields at the original resolution.

    Args:
        arr: The array for which the tangent vector field is sought.
        primitive_lattice_vectors: Define the unit cell coordinates.
        vector_fn: Function used to generate the vector field.
        normalize_fn: Function used to normalize the vector field.
        resize_max_dim: Determines the size of the array for which the tangent
            vector field is computed; `arr` is resized so that it has a maximum
            size of `vector_arr_size` along any dimension.
        resize_method: Method used in scaling `arr` prior to calculating the
            tangent vector field.

    Returns:
        The normalized vector field.
    """

    shape = arr.shape
    resize_factor = resize_max_dim / max(shape[-2:])

    if resize_factor < 1:
        assert any(d > resize_max_dim for d in shape[-2:])
        coarse_shape = shape[:-2] + tuple(
            [int(onp.ceil(d * resize_factor)) for d in shape[-2:]]
        )
        assert len(coarse_shape) == arr.ndim
        arr = utils.resample(arr, shape=coarse_shape, method=resize_method)

    tu, tv = vector_fn(arr)
    assert tu.shape == tv.shape == arr.shape

    if resize_factor < 1:
        tu = jax.image.resize(tu, shape, method=resize_method)
        tv = jax.image.resize(tv, shape, method=resize_method)
        assert tu.shape == tv.shape == shape

    tx, ty = change_vector_field_basis(
        tu,
        tv,
        u=primitive_lattice_vectors.u,
        v=primitive_lattice_vectors.v,
        x=basis.X,
        y=basis.Y,
    )
    return normalize_fn(tx, ty)


def change_vector_field_basis(
    tu: jnp.ndarray,
    tv: jnp.ndarray,
    u: jnp.ndarray,
    v: jnp.ndarray,
    x: jnp.ndarray,
    y: jnp.ndarray,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Changes the basis for a vector field.

    Specifically, given a field with amplitudes `(tu, tv)` of the basis
    vectors `(u, v)`, this function computes the amplitudes for the basis
    vectors `(x, y)`.

    Args:
        tu: The amplitude of the first basis vector in the original basis.
        tv: The amplitude of the second basis vector in the original basis.
        u: The first vector of the original basis.
        v: The second vector of the original basis.
        x: The first vector of the new basis.
        y: The second vector of the new basis.

    Returns:
        The field `(tx, ty)` in the new basis.
    """
    xy = jnp.stack([x, y], axis=-1)
    uxy = jnp.linalg.solve(xy, u)
    vxy = jnp.linalg.solve(xy, v)
    tx = tu * uxy[..., 0] + tv * vxy[..., 0]
    ty = tu * uxy[..., 1] + tv * vxy[..., 1]
    return tx, ty


def compute_gradient(arr: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Computes the gradient of `arr` with respect to `x` and `y`.

    This function uses periodic boundary conditions. The `x` and `y`
    dimensions correspond to the trailing axes of `arr`.

    Args:
        arr: The array whose gradient is sought.

    Returns:
        The `(gx, gy)` gradients along the x- and y- directions.
    """
    batch_dims = arr.ndim - 2
    padding = tuple([(0, 0)] * batch_dims + [(1, 1), (1, 1)])
    arr_padded = jnp.pad(arr, padding, mode="wrap")
    gradient_x = jnp.gradient(arr_padded, axis=-2)[..., 1:-1, 1:-1]
    gradient_y = jnp.gradient(arr_padded, axis=-1)[..., 1:-1, 1:-1]
    return gradient_x, gradient_y


# -----------------------------------------------------------------------------
# Functions related to vector field normalization.
# -----------------------------------------------------------------------------


def normalize_normal(
    tx: jnp.ndarray,
    ty: jnp.ndarray,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Normalizes the tangent vector field using the "Normal" method of [2012 Liu]."""
    magnitude = utils.magnitude(tx, ty)
    magnitude_safe = jnp.where(jnp.isclose(magnitude, 0), 1, magnitude)
    tx /= magnitude_safe
    ty /= magnitude_safe
    return tx, ty


def normalize_pol(
    tx: jnp.ndarray,
    ty: jnp.ndarray,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Normalizes the tangent vector field using the "Pol" method of [2012 Liu]."""
    magnitude = utils.magnitude(tx, ty)
    max_magnitude = jnp.amax(magnitude, axis=(-2, -1), keepdims=True)
    max_magnitude_safe = jnp.where(jnp.isclose(max_magnitude, 0.0), 1.0, max_magnitude)
    tx /= max_magnitude_safe
    ty /= max_magnitude_safe
    return tx, ty


def normalize_jones(
    tx: jnp.ndarray,
    ty: jnp.ndarray,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Generates a Jones vector field following the "Jones" method of [2012 Liu]."""
    magnitude = utils.magnitude(tx, ty)
    max_magnitude = jnp.amax(magnitude, axis=(-2, -1), keepdims=True)
    max_magnitude_safe = jnp.where(jnp.isclose(max_magnitude, 0.0), 1.0, max_magnitude)
    tx /= max_magnitude_safe
    ty /= max_magnitude_safe
    magnitude = magnitude / max_magnitude_safe

    magnitude_near_zero = jnp.isclose(magnitude, 0.0)
    magnitude_safe = jnp.where(magnitude_near_zero, 1.0, magnitude)
    tx_norm = jnp.where(magnitude_near_zero, 1 / jnp.sqrt(2), tx / magnitude_safe)
    ty_norm = jnp.where(magnitude_near_zero, 1 / jnp.sqrt(2), ty / magnitude_safe)

    phi = jnp.pi / 8 * (1 + jnp.cos(jnp.pi * magnitude))
    theta = utils.angle(tx_norm + 1j * ty_norm)

    jx = jnp.exp(1j * theta) * (tx_norm * jnp.cos(phi) - ty_norm * 1j * jnp.sin(phi))
    jy = jnp.exp(1j * theta) * (ty_norm * jnp.cos(phi) + tx_norm * 1j * jnp.sin(phi))

    return jx, jy


# -----------------------------------------------------------------------------
# Functions related to vector field generation by minimizing a functional.
# -----------------------------------------------------------------------------


def tangent_field(
    arr: jnp.ndarray,
    use_jones: bool,
    optimizer: jopt.Optimizer,
    alignment_weight: float,
    smoothness_weight: float,
    steps_dim_multiple: int,
    smoothing_kernel: jnp.ndarray,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Computes a real or complex tangent vector field.

    The field is tangent to the interfaces of features in `arr`, and varies smoothly
    between interfaces. It is obtained by optimization which minimizes a functional
    of the field which favors alignment with interfaces of features in `arr` as well
    as smoothness of the field. The maximum magnitude of the computed field is `1`.

    The tangent field is complex when `use_jones` is `True`, and real otherwise. Real
    fields are suitable for normalization using methods in this module. The complex
    field obtained when `use_jones` is `True` requires no normalization.

    Args:
        arr: The array for which the tangent field is sought.
        use_jones: Specifies whether a complex Jones field or a real tangent vector
            field is sought.
        optimizer: The optimizer used to minimize the functional.
        alignment_weight: The weight of an alignment term in the functional. Larger
            values will reward alignment with interfaces of features in `arr`.
        smoothness_weight: The weight of a smoothness term in the functional. Larger
            values will reward a smoother tangent field.
        steps_dim_multiple: Controls the number of steps in the optimization. The
            number of steps is the product of `steps_dim_multiple` and the dimension
            of the largest of the two trailing axes of `arr`.
        smoothing_kernel: Kernel used to smooth `arr` prior to the computation.

    Returns:
        The tangent vector fields, `(tx, ty)`.
    """
    tx, ty, _ = _tangent_field_with_loss(
        arr=arr,
        use_jones=use_jones,
        optimizer=optimizer,
        alignment_weight=alignment_weight,
        smoothness_weight=smoothness_weight,
        steps_dim_multiple=steps_dim_multiple,
        smoothing_kernel=smoothing_kernel,
    )
    return tx, ty


def _tangent_field_with_loss(
    arr: jnp.ndarray,
    use_jones: bool,
    optimizer: jopt.Optimizer,
    alignment_weight: float,
    smoothness_weight: float,
    steps_dim_multiple: int,
    smoothing_kernel: jnp.ndarray,
) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]:
    """Returns a tangent vector field and the functional values."""
    arr = jax.lax.stop_gradient(arr)
    arr_smoothed = utils.padded_conv(arr, smoothing_kernel, padding_mode="wrap")

    gx, gy = compute_gradient(arr_smoothed)
    tx0, ty0 = gy, -gx

    # Normalize the gradient if its maximum magnitude exceeds unity.
    magnitude = utils.magnitude(tx0, ty0)
    norm = jnp.maximum(1.0, jnp.amax(magnitude, axis=(-2, -1), keepdims=True))
    tx0 /= norm
    ty0 /= norm

    # If the permittivity is uniform, the vector field `(tx0, ty0)` will be
    # zero and the optimization will yield `nan`. Make a dummy vector field
    # to avoid this.
    permittivity_gradient_is_zero = jnp.all(
        jnp.isclose(tx0, 0.0), axis=(-2, -1), keepdims=True
    ) & jnp.all(jnp.isclose(ty0, 0.0), axis=(-2, -1), keepdims=True)
    tx0 = jnp.where(permittivity_gradient_is_zero, jnp.ones_like(tx0), tx0)
    ty0 = jnp.where(permittivity_gradient_is_zero, jnp.ones_like(ty0), ty0)

    # Remove the average phase of `(tx0, ty0)`, ensuring it is mostly real.
    tx0, ty0 = _remove_mean_phase(tx0, ty0)

    # Obtain initial `(tx, ty)` by taking the real part of `(tx0, ty0)` and
    # normalizing, so that all nonzero elements have magnitude `1`. This
    # ensures that the initial `(tx, ty)` is well aligned with `(tx0, ty0)`.
    tx, ty = normalize_normal(tx0.real, ty0.real)

    # If the Jones field is sought, transform `(tx, ty)` into a Jones field.
    if use_jones:
        tx, ty = normalize_jones(tx, ty)

    loss_fn: Callable[[jnp.ndarray, jnp.ndarray], jnp.ndarray] = functools.partial(
        _field_loss,
        tx0=tx0,
        ty0=ty0,
        alignment_weight=alignment_weight,
        smoothness_weight=smoothness_weight,
    )
    return _optimize_tangent_field(
        tx=tx,
        ty=ty,
        loss_fn=loss_fn,
        optimizer=optimizer,
        steps_dim_multiple=steps_dim_multiple,
    )


def _optimize_tangent_field(
    tx: jnp.ndarray,
    ty: jnp.ndarray,
    loss_fn: Callable[[jnp.ndarray, jnp.ndarray], jnp.ndarray],
    optimizer: jopt.Optimizer,
    steps_dim_multiple: int,
) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]:
    """Optimizes `tx` and `ty` by minimizing a functional.

    Args:
        tx: The initial `tx`.
        ty: The initial `ty`.
        loss_fn: The loss function to be minimized.
        optimizer: The optimizer used to minimize the functional.
        steps_dim_multiple: Controls the number of steps in the optimization. The
            number of steps is the product of `steps_dim_multiple` and the dimension
            of the largest of the two trailing axes of `tx`.

    Returns:
        The tangent vector fields and array giving the functional values throughout
        the optimization, `(tx, ty, values)`.
    """

    def _step_fn(step_state, dummy_x):
        del dummy_x
        step, state = step_state
        tx, ty = optimizer.params_fn(state)
        value, grads = jax.value_and_grad(loss_fn, argnums=(0, 1))(tx, ty)
        conj_grads = jax.tree_util.tree_map(jnp.conj, grads)
        state = optimizer.update_fn(step, conj_grads, state)
        state = _clip_magnitude_state(state, max_magnitude=1.0)
        return (step + 1, state), value

    num_steps = steps_dim_multiple * max(tx.shape[-2:])
    state = optimizer.init_fn((tx, ty))
    (_, state), values = jax.lax.scan(_step_fn, (0, state), xs=None, length=num_steps)
    tx, ty = optimizer.params_fn(state)
    return tx, ty, values


def _clip_magnitude_state(
    state: jopt.OptimizerState,
    max_magnitude: float,
) -> jopt.OptimizerState:
    """Extracts parameters from `state`, clips their magnitude, and repacks the state."""
    leaves, treedef = jax.tree_util.tree_flatten(state)
    assert len(leaves) % 2 == 0
    idx_tx = 0
    idx_ty = len(leaves) // 2
    tx = leaves[idx_tx]
    ty = leaves[idx_ty]
    assert tx.shape == ty.shape
    tx, ty = _clip_magnitude(tx, ty, max_magnitude)
    leaves = list(leaves)
    leaves[idx_tx] = tx
    leaves[idx_ty] = ty
    return jax.tree_util.tree_unflatten(treedef, leaves)


def _field_loss(
    tx: jnp.ndarray,
    ty: jnp.ndarray,
    tx0: jnp.ndarray,
    ty0: jnp.ndarray,
    alignment_weight: float,
    smoothness_weight: float,
) -> jnp.ndarray:
    """Returns the tangent loss.

    The tangent loss is minimized when `(tx, ty)` is aligned with
    `(tx0, ty0)` and is smooth.

    Args:
        tx: The x-component of the field to be optimized.
        ty: The y-component of the field to be optimized.
        tx0: The x-component of the target field.
        ty0: The y-component of the target field.
        alignment_weight: Determines the weight of term rewarding alignement of
            `(tx, ty)` with `(tx0, ty0)`.
        smoothness_weight: Determines the weight of the smoothness of `(tx, ty)`.

    Returns:
        The loss value.
    """
    self_alignment_loss = alignment_weight * _self_alignment_loss(tx, ty, tx0, ty0)
    gx = jnp.sum(jnp.abs(jnp.asarray(_forward_difference_gradient(tx))) ** 2)
    gy = jnp.sum(jnp.abs(jnp.asarray(_forward_difference_gradient(ty))) ** 2)
    smoothness_loss = smoothness_weight * (gx + gy)
    return self_alignment_loss + smoothness_loss


def _forward_difference_gradient(arr: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Computes the gradient by forward finite difference."""
    gx = jnp.roll(arr, axis=-2, shift=-1) - arr
    gy = jnp.roll(arr, axis=-1, shift=-1) - arr
    return gx, gy


def _self_alignment_loss(
    tx: jnp.ndarray,
    ty: jnp.ndarray,
    tx0: jnp.ndarray,
    ty0: jnp.ndarray,
) -> jnp.ndarray:
    """Returns a loss associated with alignment of `(tx, ty)` and `(tx0, ty0)`.

    Args:
        tx: The x-component of the field to be optimized.
        ty: The y-component of the field to be optimized.
        tx0: The x-component of the target field.
        ty0: The y-component of the target field.

    Returns:
        The loss value.
    """
    alignment_loss = -jnp.abs(tx * jnp.conj(tx0) + ty * jnp.conj(ty0))
    magnitude_loss = jax.nn.relu(utils.magnitude(tx, ty) - 1) ** 2
    return jnp.sum(alignment_loss + magnitude_loss)


def _clip_magnitude(
    ax: jnp.ndarray,
    ay: jnp.ndarray,
    max_magnitude: float,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Clips `(ax, ay)` to have a maximum magnitude."""
    magnitude = utils.magnitude(ax, ay)
    magnitude_safe = jnp.where(jnp.isclose(magnitude, 0.0), 1.0, magnitude)
    norm_safe = jnp.where(
        magnitude < max_magnitude,
        1.0,
        max_magnitude / magnitude_safe,
    )
    return ax * norm_safe, ay * norm_safe


def _remove_mean_phase(
    ax: jnp.ndarray, ay: jnp.ndarray
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Removes the average phase of `(ax, ay)`."""
    magnitude = utils.magnitude(ax, ay)
    magnitude_squared_safe = jnp.where(jnp.isclose(magnitude, 0.0), 1.0, magnitude**2)
    phase = (
        utils.angle(ax) * jnp.abs(ax) ** 2 + utils.angle(ay) * jnp.abs(ay) ** 2
    ) / magnitude_squared_safe
    mean_phase = jnp.mean(phase, axis=(-2, -1), keepdims=True)
    norm = jnp.exp(1j * mean_phase)
    return ax / norm, ay / norm


# -----------------------------------------------------------------------------
# Library of available vector field generating schemes.
# -----------------------------------------------------------------------------


VectorFn = Callable[
    [jnp.ndarray, basis.LatticeVectors], Tuple[jnp.ndarray, jnp.ndarray]
]

JONES_DIRECT: str = "jones_direct"
JONES: str = "jones"
NORMAL: str = "normal"
POL: str = "pol"

OPTIMIZER = jopt.momentum(step_size=0.2, mass=0.8)
ALIGNMENT_WEIGHT = 1.0
SMOOTHNESS_WEIGHT = 1.0
STEPS_DIM_MULTIPLE = 10.0
SMOOTHING_KERNEL = utils.gaussian_kernel(shape=(9, 9), fwhm=3.0)
RESIZE_MAX_DIM = 140
RESIZE_METHOD = jax.image.ResizeMethod.CUBIC


VECTOR_FIELD_SCHEMES: Dict[str, VectorFn] = {
    JONES_DIRECT: functools.partial(
        normalized_vector_field,
        vector_fn=functools.partial(
            tangent_field,
            use_jones=True,  # Directly compute Jones field.
            optimizer=OPTIMIZER,
            alignment_weight=ALIGNMENT_WEIGHT,
            smoothness_weight=SMOOTHNESS_WEIGHT,
            steps_dim_multiple=STEPS_DIM_MULTIPLE,
            smoothing_kernel=SMOOTHING_KERNEL,
        ),
        # The `JONES_DIRECT` scheme directly produces the Jones field, which we
        # just normalize to ensure the elements have a maximum magnitude of `1`.
        normalize_fn=normalize_pol,
        resize_max_dim=RESIZE_MAX_DIM,
        resize_method=RESIZE_METHOD,
    ),
    JONES: functools.partial(
        normalized_vector_field,
        vector_fn=functools.partial(
            tangent_field,
            use_jones=False,  # Obtain Jones field by normalization.
            optimizer=OPTIMIZER,
            alignment_weight=ALIGNMENT_WEIGHT,
            smoothness_weight=SMOOTHNESS_WEIGHT,
            steps_dim_multiple=STEPS_DIM_MULTIPLE,
            smoothing_kernel=SMOOTHING_KERNEL,
        ),
        normalize_fn=normalize_jones,
        resize_max_dim=RESIZE_MAX_DIM,
        resize_method=RESIZE_METHOD,
    ),
    NORMAL: functools.partial(
        normalized_vector_field,
        vector_fn=functools.partial(
            tangent_field,
            use_jones=False,
            optimizer=OPTIMIZER,
            alignment_weight=ALIGNMENT_WEIGHT,
            smoothness_weight=SMOOTHNESS_WEIGHT,
            steps_dim_multiple=STEPS_DIM_MULTIPLE,
            smoothing_kernel=SMOOTHING_KERNEL,
        ),
        normalize_fn=normalize_normal,
        resize_max_dim=RESIZE_MAX_DIM,
        resize_method=RESIZE_METHOD,
    ),
    POL: functools.partial(
        normalized_vector_field,
        vector_fn=functools.partial(
            tangent_field,
            use_jones=False,
            optimizer=OPTIMIZER,
            alignment_weight=ALIGNMENT_WEIGHT,
            smoothness_weight=SMOOTHNESS_WEIGHT,
            steps_dim_multiple=STEPS_DIM_MULTIPLE,
            smoothing_kernel=SMOOTHING_KERNEL,
        ),
        normalize_fn=normalize_pol,
        resize_max_dim=RESIZE_MAX_DIM,
        resize_method=RESIZE_METHOD,
    ),
}



"""Tests for `examples.vector`.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

import unittest

from examples import vector_fields


class PlotFieldTest(unittest.TestCase):
    def test_generate_figure(self):
        vector_fields.plot_vector_fields(savefig=False)



"""Tests for `examples.metal_grating`.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

import unittest

import numpy as onp
import parameterized

from examples import metal_grating
from fmmax import basis, fmm


class MetalGratingTest(unittest.TestCase):
    @parameterized.parameterized.expand(
        [
            (
                fmm.Formulation.FFT,
                (-0.871915 + 0.237970j),
                (0.832493 - 0.094269j),
            ),
            (
                fmm.Formulation.JONES_DIRECT,
                (-0.871906 + 0.237973j),
                (0.901329 + 0.167032j),
            ),
        ]
    )
    def test_regression(self, formulation, expected_te, expected_tm):
        # Checks that results match reference values. This helps protect against
        # regressions in accuracy of the simulator.
        (results,) = metal_grating.convergence_study(
            approximate_num_terms=(20,),
            truncations=(basis.Truncation.CIRCULAR,),
            fmm_formulations=(formulation,),
        )
        _, _, _, rte, rtm = results
        onp.testing.assert_allclose(rte, expected_te, rtol=1e-3)
        onp.testing.assert_allclose(rtm, expected_tm, rtol=1e-3)



"""Tests for `examples.ar_coating`.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

import unittest

import numpy as onp

from examples import crystal


class CrystalDipoleTest(unittest.TestCase):
    def test_regression(self):
        (
            (ex, ey, ez),
            (hx, hy, hz),
            (x, y, z),
            (section_xy, section_xz, section_yz),
        ) = crystal.simulate_crystal_with_internal_source(
            brillouin_grid_shape=(2, 3),
            resolution_fields=0.1,
        )

        self.assertSequenceEqual(ex.shape, x.shape + z.shape + (1,))
        self.assertSequenceEqual(ex.shape, ey.shape)
        self.assertSequenceEqual(ex.shape, ez.shape)
        self.assertSequenceEqual(ex.shape, hx.shape)
        self.assertSequenceEqual(ex.shape, hy.shape)
        self.assertSequenceEqual(ex.shape, hz.shape)

        onp.testing.assert_allclose(
            onp.mean(onp.abs((ex, ey, ez))), 0.466922, rtol=1e-4
        )
        onp.testing.assert_allclose(
            onp.mean(onp.abs((hx, hy, hz))), 0.306188, rtol=1e-4
        )


class CrystalGaussianBeamTest(unittest.TestCase):
    def test_regression(self):
        (
            (ex, ey, ez),
            (hx, hy, hz),
            (x, y, z),
            (section_xy, section_xz, section_yz),
        ) = crystal.simulate_crystal_with_gaussian_beam(
            brillouin_grid_shape=(2, 3),
            resolution_fields=0.1,
        )

        self.assertSequenceEqual(ex.shape, x.shape + z.shape + (1,))
        self.assertSequenceEqual(ex.shape, ey.shape)
        self.assertSequenceEqual(ex.shape, ez.shape)
        self.assertSequenceEqual(ex.shape, hx.shape)
        self.assertSequenceEqual(ex.shape, hy.shape)
        self.assertSequenceEqual(ex.shape, hz.shape)

        onp.testing.assert_allclose(
            onp.mean(onp.abs((ex, ey, ez))), 0.227245, rtol=1e-4
        )
        onp.testing.assert_allclose(onp.mean(onp.abs((hx, hy, hz))), 0.17809, rtol=1e-4)



"""An example which sets up a polarization-sorting metasurface component.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

from typing import Any, Callable, Dict, List, Optional, Tuple

import jax
import jax.example_libraries.optimizers as jopt
import jax.numpy as jnp

from fmmax import basis, fields, fmm, layer, scattering, utils

Aux = Dict[str, Any]
Initializer = Callable[[jax.random.PRNGKeyArray, Tuple[int, int]], jnp.ndarray]
Params = Dict[str, Any]

_DEFAULT_WAVELENGTH: jnp.ndarray = jnp.asarray([0.55])


class PolarizationSorterComponent:
    """Defines a polarization sorter component.

    The polarization sorter consists of a substrate subdivided into four quadrants,
    with a patterned sorter layer. The substrate and sorter are separated by a
    spacer, and the sorter is covered with a cap. Above the cap is the ambient.
    This is illustrated below.
                               __________________________
                              /                         /|
                             /                         //|
                            /                         ///|
                           /________________________ ////|
              ambient --> |_________________________|/// |
                  cap --> |_________________________|// /|
               sorter --> |_________________________|/ / |
               spacer --> |                         | /| |
                          |_________________________|/ |/
                          |            |            |  /
            substrate --> |            |            | /
                          |____________|____________|/

    The polarization sorter is illuminated with plane waves having four different
    linear polarizations. The response of the sorter (for each linear polarization)
    consists of the reflected power and the power transmitted into each quadrant.

    The polarization sorter is parameterized by the thicknesses of all its layers
    and the density of the sorter layer.
    """

    def __init__(
        self,
        ambient_permittivity: complex = (1.0 + 0.0j) ** 2,
        ambient_thickness: float = 1.0,
        cap_permittivity: complex = (1.5 + 0.0j) ** 2,
        cap_thickness: float = 0.0,
        sorter_permittivity_solid: complex = (4.0 + 0.0j) ** 2,
        sorter_permittivity_void: complex = (1.5 + 0.0j) ** 2,
        sorter_thickness: float = 0.15,
        spacer_permittivity: complex = (1.5 + 0.0j) ** 2,
        spacer_thickness: float = 1.0,
        substrate_permittivity: complex = (4.0730 + 0.028038j) ** 2,
        substrate_thickness: float = 1.0,
        wavelength: jnp.ndarray = _DEFAULT_WAVELENGTH,
        pitch: float = 2.0,
        approximate_num_terms: int = 200,
        truncation: basis.Truncation = basis.Truncation.CIRCULAR,
        formulation: fmm.Formulation = fmm.Formulation.FFT,
        density_grid_shape: Tuple[int, int] = (128, 128),
        field_grid_shape: Tuple[int, int] = (128, 128),
        field_z_offset: float = 0.0,
        density_initializer: Initializer = lambda key, shape: jnp.ones(shape) * 0.5,
    ) -> None:
        """Initializes the `PolarizationSorterComponent`.

        Args:
            ambient_permittivity: Permittivity of the ambient.
            ambient_thickness: Thickness of the ambient layer.
            cap_permittivity: Permittivity of the cap.
            cap_thickness: Thickness of the cap layer.
            sorter_permittivity_solid: Permittivity of solid regions in the sorter.
            sorter_permittivity_void: Permittivity of void regions in the sorter.
            sorter_thickness: Thickness of the sorter layer.
            spacer_permittivity: Permittivity of the spacer.
            spacer_thickness: Thickness of the spacer layer.
            substrate_permittivity: Permittivity of the substrate.
            substrate_thickness: Thickness of the susbtrate layer.
            wavelength: Wavelength, with units consistent with thicknesses.
            pitch: Pitch for the pixel quad, with units consistent with thicknesses.
            approximate_num_terms: The approximate number of terms in the expansion.
            truncation: Specifies the truncation of terms.
            formulation: Specifies the formulation to be used.
            density_grid_shape: The shape of the density grid defining the patterning
                of the sorter layer.
            field_grid_shape: The shape of the grid for the physical grid computation.
            field_z_offset: The offset into the substrate at which the power into each
                quadrant is computed.
            density_initializer: Callable used to initialize the sorter density.
        """
        self._ambient_permittivity: jnp.ndarray = jnp.asarray(ambient_permittivity)
        self._ambient_thickness: jnp.ndarray = jnp.asarray(ambient_thickness)
        self._cap_permittivity: jnp.ndarray = jnp.asarray(cap_permittivity)
        self._cap_thickness: jnp.ndarray = jnp.asarray(cap_thickness)
        self._sorter_permittivity_solid: jnp.ndarray = jnp.asarray(
            sorter_permittivity_solid
        )
        self._sorter_permittivity_void: jnp.ndarray = jnp.asarray(
            sorter_permittivity_void
        )
        self._sorter_thickness: jnp.ndarray = jnp.asarray(sorter_thickness)
        self._spacer_permittivity: jnp.ndarray = jnp.asarray(spacer_permittivity)
        self._spacer_thickness: jnp.ndarray = jnp.asarray(spacer_thickness)
        self._substrate_permittivity: jnp.ndarray = jnp.asarray(substrate_permittivity)
        self._substrate_thickness: jnp.ndarray = jnp.asarray(substrate_thickness)
        self._wavelength: jnp.ndarray = jnp.asarray(wavelength)
        self._primitive_lattice_vectors = basis.LatticeVectors(
            u=pitch * basis.X,
            v=pitch * basis.Y,
        )
        self._expansion: basis.Expansion = basis.generate_expansion(
            primitive_lattice_vectors=self._primitive_lattice_vectors,
            approximate_num_terms=approximate_num_terms,
            truncation=truncation,
        )
        self._formulation: fmm.Formulation = formulation
        self._density_grid_shape: Tuple[int, int] = density_grid_shape
        self._field_grid_shape: Tuple[int, int] = field_grid_shape
        self._field_z_offset: jnp.ndarray = jnp.asarray(field_z_offset)
        self._density_initializer: Initializer = density_initializer

    def init(self, key: jax.random.PRNGKeyArray) -> Params:
        """Returns initial parameters for the polarization sorter."""
        return {
            "primitive_lattice_vectors": self._primitive_lattice_vectors,
            "layers": {
                "ambient": {"thickness": self._ambient_thickness},
                "cap": {"thickness": self._cap_thickness},
                "sorter": {
                    "thickness": self._sorter_thickness,
                    "density": self._density_initializer(key, self._density_grid_shape),
                },
                "spacer": {"thickness": self._spacer_thickness},
                "substrate": {"thickness": self._substrate_thickness},
            },
        }

    def response(
        self,
        params: Params,
        wavelength: Optional[jnp.ndarray] = None,
        expansion: Optional[basis.Expansion] = None,
        formulation: Optional[fmm.Formulation] = None,
        field_grid_shape: Optional[Tuple[int, int]] = None,
        field_z_offset: Optional[jnp.ndarray] = None,
    ) -> Tuple[jnp.ndarray, Aux]:
        """Computes the response of the sorter defined by `params`.

        Simulation parameters may optionally be provided, to override the default
        values of the `PolarizationSorterComponent`.

        Args:
            params: Parameters for the sorter, e.g. layer thicknesses.
            wavelength: Wavelength, with units consistent with thicknesses.
            expansion: The field expansion used for the calculation.
            formulation: Specifies the formulation to be used.
            field_grid_shape: The shape of the grid for the physical grid computation.
            field_z_offset: The offset into the substrate at which the power into each
                quadrant is computed.

        Returns:
            A `(response, aux)` tuple, where the response is an array giving the
            reflected power and power transmitted into each of the quadrants, for
            each incident linear polarization. `aux` is a dictionary containing the
            grid and fields at the surface of the substrate.
        """

        if wavelength is None:
            wavelength = self._wavelength
        if expansion is None:
            expansion = self._expansion
        if formulation is None:
            formulation = self._formulation
        if field_grid_shape is None:
            field_grid_shape = self._field_grid_shape
        if field_z_offset is None:
            field_z_offset = self._field_z_offset

        return _simulate_polarization_sorter(
            params=params,
            ambient_permittivity=self._ambient_permittivity,
            cap_permittivity=self._cap_permittivity,
            sorter_permittivity_solid=self._sorter_permittivity_solid,
            sorter_permittivity_void=self._sorter_permittivity_void,
            spacer_permittivity=self._spacer_permittivity,
            substrate_permittivity=self._substrate_permittivity,
            wavelength=jnp.asarray(wavelength),
            expansion=expansion,
            formulation=formulation,
            field_grid_shape=field_grid_shape,
            field_z_offset=jnp.asarray(field_z_offset),
        )


def _simulate_polarization_sorter(
    params: Params,
    ambient_permittivity: jnp.ndarray,
    cap_permittivity: jnp.ndarray,
    sorter_permittivity_solid: jnp.ndarray,
    sorter_permittivity_void: jnp.ndarray,
    spacer_permittivity: jnp.ndarray,
    substrate_permittivity: jnp.ndarray,
    wavelength: jnp.ndarray,
    expansion: basis.Expansion,
    formulation: fmm.Formulation,
    field_grid_shape: Tuple[int, int],
    field_z_offset: jnp.ndarray,
) -> Tuple[jnp.ndarray, Aux]:
    """Simulates the polarization sorter.

    Args:
        params: Parameters for the sorter, e.g. layer thicknesses.
        ambient_permittivity: Permittivity of the ambient.
        cap_permittivity: Permittivity of the cap.
        sorter_permittivity_solid: Permittivity of solid regions in the sorter.
        sorter_permittivity_void: Permittivity of void regions in the sorter.
        spacer_permittivity: Permittivity of the spacer.
        substrate_permittivity: Permittivity of the substrate.
        wavelength: Wavelength, with units consistent with thicknesses.
        expansion: The field expansion used for the calculation.
        formulation: Specifies the formulation to be used.
        field_grid_shape: The shape of the grid for the physical grid computation.
        field_z_offset: The offset into the substrate at which the power into each
            quadrant is computed.

    Returns:
        A `(response, aux)` tuple, where the response is an array giving the
        reflected power and power transmitted into each of the quadrants, for
        each incident linear polarization. `aux` is a dictionary containing the
        grid and fields at the surface of the substrate.
    """

    # Assemble the lists of layer permittivities and thicknesses.
    permittivities = [
        ambient_permittivity[jnp.newaxis, jnp.newaxis],
        cap_permittivity[jnp.newaxis, jnp.newaxis],
        utils.interpolate_permittivity(
            permittivity_solid=sorter_permittivity_solid,
            permittivity_void=sorter_permittivity_void,
            density=params["layers"]["sorter"]["density"],
        ),
        spacer_permittivity[jnp.newaxis, jnp.newaxis],
        substrate_permittivity[jnp.newaxis, jnp.newaxis],
    ]

    thicknesses = [
        params["layers"]["ambient"]["thickness"],
        params["layers"]["cap"]["thickness"],
        params["layers"]["sorter"]["thickness"],
        params["layers"]["spacer"]["thickness"],
        params["layers"]["substrate"]["thickness"],
    ]

    # Perform the eigensolve for each layer in the stack.
    layer_solve_results = [
        layer.eigensolve_isotropic_media(
            wavelength=wavelength,
            in_plane_wavevector=jnp.zeros((2,)),
            primitive_lattice_vectors=params["primitive_lattice_vectors"],
            permittivity=p,
            expansion=expansion,
            formulation=formulation,
        )
        for p in permittivities
    ]

    s_matrix = scattering.stack_s_matrix(layer_solve_results, thicknesses)

    # We consider incident plane waves with four different linear polarizations:
    # TE, TM, (TE + TM) / sqrt(2), and (TE - TM) / sqrt(2). These defined by the
    # forward going wave amplitudes in layer `0`, i.e. the ambeint.
    n = expansion.num_terms
    fwd_amplitude_0_start = jnp.zeros((1, 2 * n, 4), dtype=complex)
    fwd_amplitude_0_start = fwd_amplitude_0_start.at[0, 0, 0].set(1)
    fwd_amplitude_0_start = fwd_amplitude_0_start.at[0, 0, 1].set(1 / jnp.sqrt(2))
    fwd_amplitude_0_start = fwd_amplitude_0_start.at[0, n, 1].set(1 / jnp.sqrt(2))
    fwd_amplitude_0_start = fwd_amplitude_0_start.at[0, 0, 2].set(1 / jnp.sqrt(2))
    fwd_amplitude_0_start = fwd_amplitude_0_start.at[0, n, 2].set(-1 / jnp.sqrt(2))
    fwd_amplitude_0_start = fwd_amplitude_0_start.at[0, n, 3].set(1)

    # The backward going wae amplitudes in layer `N`, i.e. the substrate.
    bwd_amplitude_N_end = jnp.zeros_like(fwd_amplitude_0_start)

    # Compute wave amplitudes colocated at the start of the ambient.
    bwd_amplitude_0_end = s_matrix.s21 @ fwd_amplitude_0_start
    fwd_amplitude_0_start, bwd_amplitude_0_start = fields.colocate_amplitudes(
        fwd_amplitude_0_start,
        bwd_amplitude_0_end,
        z_offset=jnp.zeros(()),  # At the start of the ambient.
        layer_solve_result=layer_solve_results[0],
        layer_thickness=params["layers"]["ambient"]["thickness"],
    )
    sz_fwd_0, sz_bwd_0 = fields.amplitude_poynting_flux(
        fwd_amplitude_0_start, bwd_amplitude_0_start, layer_solve_results[0]
    )
    sz_fwd_ambient_sum = jnp.sum(jnp.abs(sz_fwd_0), axis=-2)
    sz_bwd_ambient_sum = jnp.sum(jnp.abs(sz_bwd_0), axis=-2)
    reflection = jnp.abs(sz_bwd_ambient_sum) / jnp.abs(sz_fwd_ambient_sum)

    # Compute the forward-going and backward-going wave amplitudes in the substrate.
    fwd_amplitude_N_start = s_matrix.s11 @ fwd_amplitude_0_start
    fwd_amplitude_N_offset, bwd_amplitude_N_offset = fields.colocate_amplitudes(
        fwd_amplitude_N_start,
        bwd_amplitude_N_end,
        z_offset=field_z_offset,
        layer_solve_result=layer_solve_results[-1],
        layer_thickness=params["layers"]["substrate"]["thickness"],
    )

    # Compute the Fourier field coefficients from the wave amplitudes in the substrate.
    ef, hf = fields.fields_from_wave_amplitudes(
        fwd_amplitude_N_offset,
        bwd_amplitude_N_offset,
        layer_solve_result=layer_solve_results[-1],
    )

    # Compute the physical fields on the real-space grid in the substrate.
    (ex, ey, ez), (hx, hy, hz), (x, y) = fields.fields_on_grid(
        electric_field=ef,
        magnetic_field=hf,
        layer_solve_result=layer_solve_results[-1],
        shape=field_grid_shape,
        num_unit_cells=(1, 1),
    )
    assert ex.shape == wavelength.shape + field_grid_shape + (4,)

    # Compute the Poynting flux on the real-space grid in the substrate.
    sz = _time_average_z_poynting_flux((ex, ey, ez), (hx, hy, hz))

    # Create masks for the four quadrants.
    mask = jnp.zeros(field_grid_shape + (1, 4))
    xdim = field_grid_shape[0] // 2
    ydim = field_grid_shape[1] // 2
    mask = mask.at[:xdim, :ydim, 0, 0].set(1)
    mask = mask.at[:xdim, ydim:, 0, 1].set(1)
    mask = mask.at[xdim:, :ydim, 0, 2].set(1)
    mask = mask.at[xdim:, ydim:, 0, 3].set(1)

    # Use the mask to compute the time average Poynting flux into each quadrant. The axes for
    # `quadrant_sz` are `(wavelength, excitation, quadrant)`. Normalize this using incident power.
    quadrant_sz = jnp.mean(mask * sz[..., jnp.newaxis], axis=(-4, -3))
    quadrant_sz /= sz_fwd_ambient_sum[..., jnp.newaxis]

    # The response consists of the power into each quadrant, and the reflected power. The
    # axes for `response` are `(wavelength, excitation, port)`, where the ports correspond tos
    # reflection or transmission into one of the quadrants.
    response = jnp.concatenate([reflection[..., jnp.newaxis], quadrant_sz], axis=-1)

    aux = {"x": x, "y": y, "ex": ex, "ey": ey, "ez": ez, "hx": hx, "hy": hy, "hz": hz}

    return response, aux


def _time_average_z_poynting_flux(
    electric_fields: Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray],
    magnetic_fields: Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray],
) -> jnp.ndarray:
    """Computes the time-average z-directed Poynting flux, given the physical fields."""
    ex, ey, _ = electric_fields
    hx, hy, _ = magnetic_fields
    return jnp.real(ex * jnp.conj(hy) - ey * jnp.conj(hx))


def optimize(steps: int = 1000, approximate_num_terms: int = 400) -> List[jnp.ndarray]:
    """Optimizes the density of the polarization sorter component."""

    psc: PolarizationSorterComponent = PolarizationSorterComponent(
        approximate_num_terms=approximate_num_terms
    )

    # Define the target response, which is the response for an ideal polarization
    # sorter. The target response specifies values for reflected power and transmitted
    # power into each of the four subpixels ("quadrants"), for each of four specific
    # linear polarizations.
    target_response: jnp.ndarray = jnp.asarray(
        [
            # R,  Q1,  Q2,   Q3,   Q4
            [0.0, 0.5, 0.25, 0.25, 0.0],  # TE
            [0.0, 0.25, 0.5, 0.0, 0.25],  # (TE + TM) / sqrt(2)
            [0.0, 0.25, 0.0, 0.5, 0.25],  # (TE - TM) / sqrt(2)
            [0.0, 0.0, 0.25, 0.25, 0.5],  # TM
        ]
    )

    params: Params = psc.init(jax.random.PRNGKey(0))
    density = params["layers"]["sorter"]["density"]

    opt: jopt.Optimizer = jopt.adam(0.002, b1=0.67, b2=0.9)
    opt_state = opt.init_fn(density)

    def train_step(
        step: int,
        opt_state: Any,
    ) -> Tuple[Any, jnp.ndarray, jnp.ndarray, Aux]:
        def loss_fn(
            density: jnp.ndarray,
        ) -> Tuple[jnp.ndarray, Tuple[jnp.ndarray, Aux]]:
            params["layers"]["sorter"]["density"] = density
            response, aux = psc.response(params)
            loss = jnp.sum((response - target_response) ** 2) / response.shape[0]
            return loss, (response, aux)

        (value, (response, aux)), grads = jax.value_and_grad(loss_fn, has_aux=True)(
            opt.params_fn(opt_state)
        )
        opt_state = opt.update_fn(step, grads, opt_state)

        # Clip the density to the range (0, 1). Since we are using the jax example
        # optimizer library we must unpack the state and manipulate the stored
        # parameter, and the finally repack.
        (density, *aux_leaves), treedef = jax.tree_util.tree_flatten(opt_state)
        leaves = [density] + aux_leaves
        opt_state = jax.tree_util.tree_unflatten(treedef, leaves)
        return opt_state, value, response, aux

    values = []
    for i in range(steps):
        opt_state, value, response, _ = train_step(i, opt_state)
        values.append(value)
        print(i, value)
    density = opt.params_fn(opt_state)

    # Check if the result is converged.
    params["layers"]["sorter"]["density"] = density
    for factor in (0.5, 1.0, 1.5, 2.0):
        n = int(factor * approximate_num_terms)
        response, _ = psc.response(
            params,
            expansion=basis.generate_expansion(
                primitive_lattice_vectors=params["primitive_lattice_vectors"],
                approximate_num_terms=n,
                truncation=basis.Truncation.CIRCULAR,
            ),
        )
        print(f"Response with approximate_num_terms={n}:")
        print(response)

    return values


if __name__ == "__main__":
    optimize()



"""Example which exercises automatic vector field generation.

Copyright (c) Meta Platforms, Inc. and affiliates.
"""

import jax.numpy as jnp
import matplotlib.pyplot as plt  # type: ignore[import]

from fmmax import basis, fmm, vector


def plot_vector_fields(
    dim: int = 100, interval: int = 10, savefig: bool = True
) -> None:
    """Generate a figure that compares different vector field schemes."""

    # Generate the pattern; a circular feature in a square unit cell.
    primitive_lattice_vectors = basis.LatticeVectors(u=basis.X, v=basis.Y)
    x, y = jnp.meshgrid(
        jnp.linspace(-0.5, 0.5, dim),
        jnp.linspace(-0.5, 0.5, dim),
        indexing="ij",
    )
    arr = (x**2 + y**2 < 0.4**2).astype(float)

    fig = plt.figure(figsize=(7, 3))

    schemes = [f for f in fmm.Formulation if f != fmm.Formulation.FFT]

    for i, formulation in enumerate(schemes):
        tx, ty = vector.VECTOR_FIELD_SCHEMES[formulation.value](
            arr, primitive_lattice_vectors
        )

        ax = plt.subplot(1, 4, i + 1)
        _plot_vector_field(
            ax, arr, tx, ty, scale=0.7, interval=interval, title=formulation.value
        )

    plt.subplots_adjust(left=0, bottom=0, right=1, top=1, wspace=0.05)

    if savefig:
        fig.savefig("vector_fields.png", bbox_inches="tight")


# -----------------------------------------------------------------------------
# Plotting functions.
# -----------------------------------------------------------------------------


def _plot_vector_field(
    ax: plt.Axes,
    arr: jnp.ndarray,
    vx: jnp.ndarray,
    vy: jnp.ndarray,
    scale: float = 0.7,
    interval: int = 1,
    lw: float = 0.25,
    title: str = "",
) -> None:
    """Plots an overlay of `arr` with the vector field."""
    im = ax.imshow(arr, cmap="gray")
    im.set_clim([-5 * (jnp.amax(arr) - jnp.amin(arr)), jnp.amax(arr)])

    x, y = jnp.meshgrid(*[jnp.arange(d) for d in vx.shape], indexing="ij")

    x = _sample(x, interval)
    y = _sample(y, interval)
    vx = _sample(vx, interval)
    vy = _sample(vy, interval)

    if jnp.any(jnp.iscomplex(vx)) or jnp.any(jnp.iscomplex(vy)):
        _plot_polarization_elipses(ax, x, y, vx, vy, scale=scale * interval, lw=lw)
    else:
        _plot_polarization_vectors(ax, x, y, vx, vy)
    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_title(title)


def _sample(arr: jnp.ndarray, interval: int) -> jnp.ndarray:
    """Samples every elements from `arr`, selecting every `interval` element."""
    return arr[(interval // 2) :: interval, (interval // 2) :: interval]


def _plot_polarization_vectors(
    ax: plt.Axes, x: jnp.ndarray, y: jnp.ndarray, vx: jnp.ndarray, vy: jnp.ndarray
) -> None:
    """Plots a real vector field."""
    ax.quiver(y, x, -vy, vx)


def _plot_polarization_elipses(
    ax: plt.Axes,
    x: jnp.ndarray,
    y: jnp.ndarray,
    vx: jnp.ndarray,
    vy: jnp.ndarray,
    scale: float,
    lw: float,
) -> None:
    """Plots a complex vector field."""
    t = jnp.concatenate([jnp.linspace(0, 2 * jnp.pi), jnp.asarray((jnp.nan,))])[
        jnp.newaxis, :
    ]
    x = x.flatten()[:, jnp.newaxis]
    y = y.flatten()[:, jnp.newaxis]
    vx = vx.flatten()[:, jnp.newaxis]
    vy = vy.flatten()[:, jnp.newaxis]
    xplot = x + jnp.real(vx * jnp.exp(1j * t)) * scale / 2
    yplot = y + jnp.real(vy * jnp.exp(1j * t)) * scale / 2
    ax.plot(yplot.flatten(), xplot.flatten(), lw=lw, color="k")


if __name__ == "__main__":
    plot_vector_fields()
